# -*- coding: utf-8 -*-
"""
lories.data.manager
~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
import os
import signal
import time
from collections.abc import Callable
from concurrent import futures
from concurrent.futures import Future, ThreadPoolExecutor, TimeoutError
from functools import partial
from threading import Event, Thread
from typing import Any, Dict, Mapping, Optional, Type

import pandas as pd
import pytz as tz
from lories._core import _Context, _DataManager  # noqa
from lories.components import Component, ComponentContext
from lories.connectors import Connector, ConnectorContext, ConnectorError
from lories.connectors.tasks import CheckTask, ConnectTask, LogTask, ReadTask, WriteTask
from lories.core.activator import Activator
from lories.core.configs import ConfigurationError, Configurations
from lories.core.register import Registrator, RegistratorContext
from lories.core.typing import ChannelsArgument, Timestamp
from lories.data.channels import Channel, ChannelConnector, ChannelConverter, Channels, ChannelState
from lories.data.context import DataContext
from lories.data.converters import ConverterContext
from lories.data.databases import Database, Databases
from lories.data.listeners import ListenerContext
from lories.data.replication import Replication
from lories.data.retention import Retention
from lories.util import floor_date, parse_type, to_bool, to_timedelta, validate_key

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


# noinspection PyProtectedMember
class DataManager(_DataManager, DataContext, Activator):
    _converters: ConverterContext
    _connectors: ConnectorContext
    _components: ComponentContext

    _listeners: ListenerContext

    _executor: ThreadPoolExecutor
    __runner: Thread
    __interrupt: Event

    _interval: int

    def __init__(self, configs: Configurations, name: str, **kwargs) -> None:
        super().__init__(configs=configs, key=validate_key(name), name=name, **kwargs)
        self.__interrupt = Event()
        self.__interrupt.set()

        self._converters = ConverterContext(self)
        self._connectors = ConnectorContext(self)
        self._components = ComponentContext(self)
        self._listeners = ListenerContext(self)
        self._executor = ThreadPoolExecutor(
            thread_name_prefix=self.name,
            max_workers=max(int((os.cpu_count() or 1) / 2), 1),
        )
        self.__runner = Thread(name=self.name, target=self.run)

        signal.signal(signal.SIGINT, self.interrupt)
        signal.signal(signal.SIGTERM, self.deactivate)

    # noinspection PyArgumentList
    def __contains__(self, item: str | Channel | Connector | Component) -> bool:
        channels = _Context.__getattribute__(self, f"{_Context.__name__}__map")
        if isinstance(item, str):
            return item in channels.keys()
        if isinstance(item, Channel):
            return item in channels.values()
        if isinstance(item, Connector):
            return item in self._connectors.values()
        if isinstance(item, Component):
            return item in self._components.values()
        return False

    # noinspection PyShadowingBuiltins
    def _create(self, id: str, key: str, type: Type, **configs: Any) -> Channel:
        # noinspection PyShadowingBuiltins
        def build_args(
            registrator_context: RegistratorContext,
            registrator_type: str,
            name: Optional[str] = None,
        ) -> Dict[str, Any]:
            if name is None:
                name = registrator_type
            registrator_configs = configs.pop(name, None)
            if registrator_configs is None:
                return {registrator_type: None}
            if isinstance(registrator_configs, str):
                registrator_configs = {registrator_type: registrator_configs}
            elif not isinstance(registrator_configs, Mapping):
                raise ConfigurationError(f"Invalid channel {name} type: " + str(registrator_configs))
            elif registrator_type not in registrator_configs:
                return {registrator_type: None}

            registrator_id = registrator_configs.pop(registrator_type)
            if registrator_id is not None and "." not in registrator_id:
                registrator_path = id.split(".")
                for i in reversed(range(1, len(registrator_path))):
                    _registrator_id = ".".join([*registrator_path[:i], registrator_id])
                    if _registrator_id in registrator_context.keys():
                        registrator_id = _registrator_id
                        break
            registrator = registrator_context.get(registrator_id, None) if registrator_id else None
            return {registrator_type: registrator, **registrator_configs}

        if "converter" not in configs:
            converter = ChannelConverter(self._converters.get_by_dtype(parse_type(type)))
        else:
            converter = ChannelConverter(**build_args(self._converters, "converter"))
        connector = ChannelConnector(**build_args(self._connectors, "connector"))
        logger = ChannelConnector(**build_args(self._connectors, "connector", "logger"))

        return Channel(
            id=id, key=key, type=type, context=self, converter=converter, connector=connector, logger=logger, **configs
        )

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._interval = configs.get_int("interval", default=1)

    def _at_configure(self, configs: Configurations) -> None:
        super()._at_configure(configs)
        self._load(self, configs, sort=False)

        self._converters.load(configure=False, sort=False)
        self._converters.configure()

        self._connectors.load(configure=False, sort=False)
        self._connectors.configure()

        self._components.load(configure=False, sort=False)
        self._components.configure()

    def _on_configure(self, configs: Configurations) -> None:
        super()._on_configure(configs)
        self._converters.sort()
        self._connectors.sort()
        self._components.sort()
        self.sort()

    # noinspection PyShadowingBuiltins
    def activate(self, filter: Optional[Callable[[Registrator], bool]] = None) -> None:
        super().activate()
        self._connect(*self._connectors.filter(_filter(filter)))
        self._activate(*self._components.filter(_filter(filter)))

    def _activate(self, *component: Component) -> None:
        for component in component:
            if not component.is_enabled():
                self._logger.debug(
                    f"Skipping to activate disabled {type(component).__name__} '{component.name}': {component.id}"
                )
                continue
            self.__activate(component)

    def __activate(self, component: Component) -> None:
        self._logger.debug(f"Activating {type(component).__name__} '{component.name}': {component.id}")
        component.activate()

        self._logger.info(f"Activated {type(component).__name__} '{component.name}': {component.id}")

    # noinspection PyShadowingBuiltins
    def connect(
        self,
        filter: Optional[Callable[[Registrator], bool]] = None,
        channels: Optional[Channels] = None,
        timeout: Optional[float] = None,
    ) -> None:
        self._connect(*self._connectors.filter(_filter(filter)), channels=channels, timeout=timeout)

    def _connect(
        self,
        *connectors: Connector,
        channels: Optional[Channels] = None,
        timeout: Optional[float] = None,
        force: bool = False,
    ) -> None:
        connect_futures = {}
        for connector in connectors:
            if not connector.is_enabled():
                self._logger.debug(
                    f"Skipping to connect disabled {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            if connector._is_connected():
                self._logger.debug(
                    f"Skipping already connected {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            if not connector._is_connectable() and not force:
                self._logger.debug(
                    f"Skipping not connectable {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            connect_task = self.__connect(connector, channels)
            connect_future = self._executor.submit(connect_task)
            connect_futures[connect_future] = connect_task

        self.__connect_futures(connect_futures, timeout)

    def __connect(self, connector: Connector, channels: Optional[Channels] = None) -> ConnectTask:
        self._logger.debug(f"Connecting {type(connector).__name__} '{connector.name}': {connector.id}")
        if channels is None:
            channels = self.channels.filter(lambda c: c.has_connector(connector.id))
            channels.update(self.channels.filter(lambda c: c.has_logger(connector.id)).apply(lambda c: c.from_logger()))

        return ConnectTask(connector, channels)

    def __connect_futures(
        self,
        tasks: Dict[Future, ConnectTask],
        timeout: Optional[float] = None,
    ) -> None:
        try:
            for future in futures.as_completed(tasks, timeout=timeout):
                tasks.pop(future)
                self.__connect_callback(future)

        except TimeoutError:
            for future, task in tasks.items():
                self._logger.warning(f"Timed out opening connector '{task.connector.id}' after {timeout} seconds")
                future.cancel()

    def __connect_callback(self, future: Future) -> None:
        try:
            connector = future.result()
            self._logger.info(f"Connected {type(connector).__name__} '{connector.name}': {connector.id}")

        except ConnectorError as e:
            self._logger.warning(f"Failed opening connector '{e.connector.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)

    # noinspection PyShadowingBuiltins
    def reconnect(
        self,
        filter: Optional[Callable[[Registrator], bool]] = None,
    ) -> None:
        self._reconnect(*self._connectors.filter(_filter(filter)))

    def _reconnect(self, *connectors: Connector) -> None:
        for connector in connectors:
            if not connector.is_enabled():
                self._logger.debug(
                    f"Skipping reconnecting disabled {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            if not connector._is_connected() and connector._connected:
                # Connection aborted and not yet disconnected properly
                self._disconnect(connector)
                continue

            connect_task = self.__connect(connector)
            connect_future = self._executor.submit(connect_task)
            connect_future.add_done_callback(self.__connect_callback)

    # noinspection PyShadowingBuiltins
    def disconnect(
        self,
        filter: Optional[Callable[[Registrator], bool]] = None,
    ) -> None:
        self._disconnect(*self._connectors.filter(_filter(filter)))

    def _disconnect(self, *connectors: Connector) -> None:
        for connector in reversed(connectors):
            if not connector._is_connected():
                self._logger.debug(
                    f"Skipping to disconnect unconnected {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue
            self.__disconnect(connector)

    def __disconnect(self, connector: Connector) -> None:
        try:
            self._logger.debug(f"Disconnecting {type(connector).__name__} '{connector.name}': {connector.id}")
            connector.set_channels(ChannelState.DISCONNECTING)
            connector.disconnect()

            self._logger.info(f"Disconnected {type(connector).__name__} '{connector.name}': {connector.id}")

        except Exception as e:
            self._logger.warning(f"Failed closing connector '{connector.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)
        finally:
            connector.set_channels(ChannelState.DISCONNECTED)

    # noinspection PyShadowingBuiltins
    def deactivate(self, *_, filter: Optional[Callable[[Registrator], bool]] = None) -> None:
        self.interrupt()
        super().deactivate()
        self._deactivate(*self._components.filter(_filter(filter)))
        self._disconnect(*self._connectors.filter(_filter(filter)))

    def _deactivate(self, *components: Component) -> None:
        for component in reversed(list(components)):
            if not component.is_active():
                self._logger.debug(
                    f"Skipping to deactivate already deactivated {type(component).__name__} '{component.name}': "
                    f"{component.id}"
                )
                return
            self.__deactivate(component)

    def __deactivate(self, component: Component) -> None:
        if not component.is_active():
            self._logger.debug(
                f"Skipping to deactivate already deactivated {type(component).__name__} '{component.name}': "
                f"{component.id}"
            )
            return
        try:
            self._logger.debug(f"Deactivating {type(component).__name__} '{component.name}': {component.id}")
            component.deactivate()

            self._logger.info(f"Deactivated {type(component).__name__} '{component.name}': {component.id}")

        except Exception as e:
            self._logger.warning(f"Failed deactivating component '{component.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)

    def interrupt(self, *_) -> None:
        self.__interrupt.set()

        # FIXME: Add cancel_futures argument again, once Python >= 3.9 is a requirement
        self._executor.shutdown(wait=True)  # , cancel_futures=True)
        if self.__runner.is_alive():
            self.__runner.join()

    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        channels: Optional[ChannelsArgument] = None,
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        self._listeners.register(function, self._filter_by_args(channels), how=how, unique=unique)

    @property
    def converters(self) -> ConverterContext:
        return self._converters

    @property
    def connectors(self) -> ConnectorContext:
        return self._connectors

    @property
    def components(self) -> ComponentContext:
        return self._components

    @property
    def listeners(self) -> ListenerContext:
        return self._listeners

    def notify(
        self,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
    ) -> None:
        channels = self._filter_by_args(channels)
        now = pd.Timestamp.now(tz.UTC)

        def _submit_listeners(_timeout: float) -> bool:
            _futures = []
            with self.listeners:
                for _listener in self.listeners.notify(*channels):
                    _future = self._executor.submit(_listener, now)
                    _future.add_done_callback(self._notify_callback)
                    _futures.append(_future)
            if len(_futures) > 0:
                futures.wait(_futures, timeout=_timeout)
                return True
            return False

        while _submit_listeners(timeout):
            if timeout is not None:
                timeout -= (pd.Timestamp.now(tz.UTC) - now).total_seconds()
                if timeout <= 0:
                    break

    # noinspection PyUnresolvedReferences
    def _notify_callback(self, future: Future) -> None:
        exception = future.exception()
        if exception is not None:
            listener = exception.listener
            self._logger.warning(f"Failed notifying listener '{listener.id}': {str(exception)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(exception)

    def start(self, wait: bool = True) -> None:
        self._logger.info(f"Starting {type(self).__name__}: {self.name}")
        self.__interrupt.clear()
        self.__runner.start()
        if wait:
            self.__runner.join()

    # noinspection PyShadowingBuiltins, PyProtectedMember
    def run(self, **kwargs) -> None:
        now = pd.Timestamp.now(tz.UTC)

        channels = self.channels.filter(lambda c: self.__is_reading(c, now))
        if len(channels) > 0:
            self.read(channels, inplace=True, **kwargs)

        interval = f"{self._interval}s"
        _sleep(interval)

        while not self.__interrupt.is_set():
            try:
                now = pd.Timestamp.now(tz.UTC)

                self.__read(now, timeout=self._interval / 4)

                self.reconnect(lambda c: c._is_reconnectable())
                self.notify(timeout=self._interval / 4)
                self.log()

                _sleep(interval, self.__interrupt.wait)

            except KeyboardInterrupt:
                self.interrupt()
                break

        self.notify()
        self.log()

    # noinspection PyShadowingBuiltins
    def has_logged(
        self,
        channels: Optional[ChannelsArgument] = None,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> bool:
        channels = self._filter_by_args(channels)

        check_futures = {}
        for id, connector in self.connectors.items():
            if not connector._is_connected():
                continue

            def has_database(channel: Channel) -> bool:
                return channel.has_logger(id) and channel.logger.is_database()

            check_channels = channels.filter(has_database).apply(lambda c: c.from_logger())
            if len(check_channels) == 0:
                continue

            check_task = CheckTask(connector, check_channels)
            check_future = self._executor.submit(check_task, start=start, end=end)
            check_futures[check_future] = check_task

        check_results = []
        try:
            for check_future in futures.as_completed(check_futures, timeout=timeout):
                check_task = check_futures.pop(check_future)
                try:
                    check_exists = check_future.result()
                    check_results.append(check_exists)

                except ConnectorError as e:
                    self._logger.warning(f"Failed checking connector '{check_task.connector.id}': {str(e)}")
                    if self._logger.getEffectiveLevel() <= logging.DEBUG:
                        self._logger.exception(e)

                    check_results.append(False)

        except TimeoutError:
            for check_future, check_task in check_futures.items():
                self._logger.warning(
                    f"Timed out checking connector '{check_task.connector.id}' after {timeout} seconds"
                )
                check_future.cancel()
                check_results.append(False)

        if len(check_results) == 0:
            return False
        return all(check_results)

    # noinspection PyShadowingBuiltins
    def read_logged(
        self,
        channels: Optional[ChannelsArgument] = None,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame:
        channels = self._filter_by_args(channels)

        read_futures = {}
        for id, connector in self.connectors.items():
            if not connector._is_connected():
                continue

            def has_database(channel: Channel) -> bool:
                return channel.has_logger(id) and channel.logger.is_database()

            read_channels = channels.filter(has_database).apply(lambda c: c.from_logger())
            if len(read_channels) == 0:
                continue

            read_task = ReadTask(connector, read_channels)
            read_future = self._executor.submit(read_task, start=start, end=end)
            read_futures[read_future] = read_task

        return self._read_futures(read_futures, timeout)

    # noinspection PyShadowingBuiltins, PyTypeChecker
    def read(
        self,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
        inplace: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        channels = self._filter_by_args(channels)

        read_futures = {}
        for id, connector in self.connectors.items():
            if not connector._is_connected():
                continue

            read_channels = channels.filter(lambda c: c.has_connector(id))
            if len(read_channels) == 0:
                continue

            read_task = ReadTask(connector, read_channels)
            read_future = self._executor.submit(read_task, inplace=inplace, **kwargs)
            read_futures[read_future] = read_task

        return self._read_futures(read_futures, timeout, inplace)

    def _read_futures(
        self,
        tasks: Dict[Future, ReadTask],
        timeout: Optional[float] = None,
        inplace: bool = False,
    ) -> pd.DataFrame:
        results = []
        try:
            for future in futures.as_completed(tasks, timeout=timeout):
                task = tasks.pop(future)
                data = self._read_callback(task, future, inplace)
                if data is not None:
                    results.append(data)

        except TimeoutError:
            for future, task in tasks.items():
                self._logger.warning(f"Timed out reading connector '{task.connector.id}' after {timeout} seconds")
                future.cancel()
                if inplace:
                    channels = task.channels
                    channels.set_state(ChannelState.TIMEOUT)

        if len(results) == 0:
            return pd.DataFrame()
        results = sorted(results, key=lambda d: min(d.index))
        return pd.concat(results, axis="columns")

    def _read_callback(
        self,
        task: ReadTask,
        future: Future,
        inplace: bool = False,
    ) -> Optional[pd.DataFrame]:
        channels = task.channels
        try:
            return future.result()

        except ConnectorError as e:
            self._logger.warning(f"Failed reading connector '{task.connector.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)
            if inplace:
                channels.set_state(ChannelState.READ_ERROR)
        return None

    # noinspection PyShadowingBuiltins, PyTypeChecker
    def __read(
        self,
        timestamp: Timestamp,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> None:
        channels = self.channels.filter(lambda c: self.__is_reading(c, timestamp))
        if len(channels) < 1:
            return
        self._logger.debug(f"Reading {len(channels)} channels of application: {self.name}")

        read_futures = []
        for id, connector in self.connectors.items():
            if not connector._is_connected():
                continue

            read_channels = channels.filter(lambda c: c.has_connector(id))
            if len(read_channels) == 0:
                continue

            read_task = ReadTask(connector, read_channels)
            read_future = self._executor.submit(read_task, inplace=True, **kwargs)
            read_future.add_done_callback(partial(self._read_callback, read_task, inplace=True))
            read_futures.append(read_future)

            def update_timestamp(read_channel: Channel) -> None:
                read_channel.connector.timestamp = timestamp

            read_channels.apply(update_timestamp, inplace=True)

        futures.wait(read_futures, timeout=timeout)

    def __is_reading(self, channel: Channel, timestamp: pd.Timestamp) -> bool:
        freq = channel.freq
        if (
            freq is None
            or not channel.has_connector()
            or not self.connectors.get(channel.connector.id, False)
            or not self.connectors.get(channel.connector.id).is_connected()
        ):
            return False
        if pd.isna(channel.connector.timestamp):
            return True
        next_reading = _next(freq, channel.connector.timestamp)
        return timestamp >= next_reading

    # noinspection PyShadowingBuiltins, PyShadowingNames, PyTypeChecker
    def write(
        self,
        data: pd.DataFrame,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
        inplace: bool = False,
    ) -> None:
        channels = self._filter_by_args(channels)

        write_futures = {}
        for id, connector in self.connectors.items():
            if not connector._is_connected():
                continue

            write_channels = channels.filter(lambda c: (c.has_connector(id) and c.id in data.columns))
            if len(write_channels) == 0:
                continue

            write_channels.set_frame(data)
            write_task = WriteTask(connector, write_channels)
            write_future = self._executor.submit(write_task)
            write_futures[write_future] = write_task

        self._write_futures(write_futures, timeout)

    def _write_futures(
        self,
        tasks: Dict[Future, WriteTask | LogTask],
        timeout: Optional[float] = None,
        inplace: bool = False,
    ) -> None:
        try:
            for future in futures.as_completed(tasks, timeout=timeout):
                task = tasks.pop(future)
                self._write_callback(task, future, inplace)

        except TimeoutError:
            for future, task in tasks.items():
                self._logger.warning(f"Timed out writing connector '{task.connector.id}' after {timeout} seconds")
                future.cancel()
                if inplace:
                    channels = task.channels
                    channels.set_state(ChannelState.TIMEOUT)

    def _write_callback(
        self,
        task: WriteTask,
        future: Future,
        inplace: bool = False,
    ) -> None:
        channels = task.channels
        try:
            future.result()

        except ConnectorError as e:
            self._logger.warning(f"Failed writing connector '{task.connector.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)
            if inplace:
                channels.set_state(ChannelState.WRITE_ERROR)

    # noinspection PyShadowingBuiltins, PyTypeChecker
    def log(
        self,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
        blocking: bool = False,
        force: bool = False,
    ) -> None:
        if channels is None:
            channels = self.channels

        log_futures = {}
        for id, connector in self.connectors.items():
            if not connector._is_connected():
                continue

            def has_update(channel: Channel) -> bool:
                if force:
                    return True
                if channel.freq is None:
                    return pd.isna(channel.logger.timestamp) or channel.timestamp > channel.logger.timestamp
                if pd.isna(channel.logger.timestamp):
                    logger_timestamp = floor_date(channel.timestamp, freq=channel.freq)
                    if logger_timestamp == channel.timestamp:
                        logger_timestamp -= channel.timedelta
                    channel.logger.timestamp = logger_timestamp

                return channel.timestamp >= channel.logger.timestamp + channel.timedelta

            log_channels = channels.filter(lambda c: (c.has_logger(id) and c.is_valid() and has_update(c)))
            if len(log_channels) == 0:
                continue

            log_task = LogTask(connector, log_channels)
            log_future = self._executor.submit(log_task)
            log_futures[log_future] = log_task
            if not blocking:
                log_future.add_done_callback(partial(self._write_callback, log_task, inplace=False))

            def update_timestamp(channel: Channel) -> None:
                channel.logger.timestamp = channel.timestamp

            log_channels.apply(update_timestamp, inplace=True)

        if blocking:
            self._write_futures(log_futures, timeout, inplace=False)

    def rotate(
        self,
        channels: Optional[Channels] = None,
        full: bool = False,
        **kwargs,
    ) -> None:
        if channels is None:
            channels = self.channels

        defaults = self.configs.get_member(Retention.TYPE, defaults={})
        configs = Configurations(f"{Retention.TYPE}.conf", self.configs.dirs, defaults=defaults)
        configs._load(require=False)
        kwargs["full"] = configs.pop("full", default=full)

        databases = Databases(self, configs)
        databases.rotate(channels, **kwargs)

    def replicate(
        self,
        channels: Optional[Channels] = None,
        full: bool = False,
        force: bool = False,
        **kwargs,
    ) -> None:
        if channels is None:
            channels = self.channels.filter(lambda c: self.__is_replicating(c))

        defaults = self.configs.get_member(Replication.TYPE, defaults={})
        configs = Configurations(f"{Replication.TYPE}.conf", self.configs.dirs, defaults=defaults)
        configs._load(require=False)
        if not configs.enabled:
            self._logger.error(f"Unable to replicate for disabled configuration type '{Replication.TYPE}'")
            return
        kwargs["full"] = configs.pop("full", default=full)
        kwargs["force"] = configs.pop("force", default=force)
        kwargs.update({k: v for k, v in configs.items() if k not in configs.members})

        databases = Databases(self, configs)
        databases.replicate(channels, **kwargs)

    # noinspection PyMethodMayBeStatic
    def __is_replicating(self, channel: Channel, timestamp: Optional[Timestamp] = None) -> bool:
        replication = channel.get(Replication.TYPE, default=None)
        if not (
            replication is not None
            and "database" in replication
            and to_bool(replication.get("enabled", True))
            and channel.logger.enabled
            and isinstance(channel.logger._connector, Database)
        ):
            return False
        if timestamp is None:
            return True
        return timestamp <= floor_date(timestamp, freq=replication.get("freq", Replication.freq))


# noinspection PyShadowingBuiltins
def _sleep(freq: str, sleep: Callable = time.sleep) -> None:
    now = pd.Timestamp.now(tz.UTC)
    next = _next(freq, now)
    seconds = (next - now).total_seconds()
    sleep(seconds)


# noinspection PyShadowingBuiltins, PyShadowingNames
def _next(freq: str, now: Optional[pd.Timestamp] = None) -> pd.Timestamp:
    if now is None:
        now = pd.Timestamp.now(tz.UTC)
    next = floor_date(now, freq=freq)
    while next <= now:
        next += to_timedelta(freq)
    return next


def _filter(*filters: Optional[Callable[[Connector | Component], bool]]) -> Callable[[...], bool]:
    def _all_filters(registrator: Connector | Component) -> bool:
        return all(f(registrator) for f in filters if f is not None)

    return _all_filters



# -*- coding: utf-8 -*-
"""
lories.core.errors
~~~~~~~~~~~~~~~~~~


"""


class ResourceError(Exception):
    """
    Raise if an error occurred accessing a local resource.

    """


class ResourceUnavailableError(ResourceError):
    """
    Raise if an accessed local resource can not be found.

    """



# -*- coding: utf-8 -*-
"""
lories.connectors.csv
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import os
from typing import Mapping, Optional, Tuple

import pandas as pd
from lories.connectors import ConnectionError, Database, register_connector_type
from lories.core.configs import ConfigurationError
from lories.io import csv
from lories.typing import Configurations, Resources, Timestamp
from lories.util import ceil_date, floor_date, parse_freq


# noinspection PyShadowingBuiltins
@register_connector_type("csv")
class CsvDatabase(Database):
    _data: Optional[pd.DataFrame] = None
    _data_path: Optional[str] = None
    _data_dir: str

    index_column: str = "timestamp"
    index_type: str = "timestamp"

    override: bool = False
    slice: bool = False

    freq: str = "D"
    format: str = None
    suffix: Optional[str] = None

    decimal: str = "."
    separator: str = ","

    columns: Mapping[str, str] = {}
    pretty: bool = False

    # noinspection PyTypeChecker
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        data_dir = configs.get("dir", default=None)
        if data_dir is not None:
            if "~" in data_dir:
                data_dir = os.path.expanduser(data_dir)
            if not os.path.isabs(data_dir):
                data_dir = os.path.join(configs.dirs.data, data_dir)

        data_path = configs.get("file", default=None)
        if data_path is not None:
            if not os.path.isabs(data_path):
                if data_dir is None:
                    data_dir = configs.dirs.data
                data_path = os.path.join(data_dir, data_path)
            elif data_dir is None:
                data_dir = os.path.dirname(data_path)

        if data_dir is None:
            data_dir = configs.dirs.data

        self._data_dir = data_dir
        self._data_path = data_path

        self.index_column = configs.get("index_column", default=CsvDatabase.index_column)
        self.index_type = configs.get("index_type", default=CsvDatabase.index_type).lower()
        if self.index_type not in ["timestamp", "unix", "none", None]:
            raise ConfigurationError(f"Unknown index type: {self.index_type}")

        self.override = configs.get_bool("override", default=CsvDatabase.override)
        self.slice = configs.get_bool("slice", default=CsvDatabase.slice)

        self.freq = parse_freq(configs.get("freq", default=CsvDatabase.freq))

        format = configs.get("format", default=CsvDatabase.format)
        if format is not None:
            self.format = format
        elif self.freq == "Y":
            self.format = "%Y"
        elif self.freq == "M":
            self.format = "%Y-%m"
        elif any([self.freq.endswith("D")]):
            self.format = "%Y%m%d"
        elif any([self.freq.endswith(s) for s in ["h", "min", "s"]]):
            self.format = "%Y%m%d_%H%M%S"
        else:
            raise ConfigurationError(f"Invalid frequency: {self.freq}")

        self.suffix = configs.get("suffix", default=CsvDatabase.suffix)
        if self.suffix is not None:
            self.format += f"_{self.suffix}"

        self.decimal = configs.get("decimal", CsvDatabase.decimal)
        self.separator = configs.get("separator", CsvDatabase.separator)

        self.pretty = configs.get_bool("pretty", default=False)
        self.columns = configs.get("columns", default=CsvDatabase.columns)

    def _build_columns(self, resources: Optional[Resources] = None) -> Mapping[str, str]:
        columns = {r.id: self.columns[r.key] for r in resources if r.key in self.columns}
        if self.pretty:
            columns.update({r.id: r.full_name(unit=True) for r in self.resources if "name" in r})
            if resources is not None:
                columns.update({r.id: r.full_name(unit=True) for r in resources if "name" in r})
        else:
            columns.update({r.id: r.get("column", default=r.key) for r in self.resources})
            if resources is not None:
                columns.update({r.id: r.get("column", default=r.key) for r in resources})
        return columns

    def connect(self, resources: Resources) -> None:
        if not os.path.isdir(self._data_dir):
            os.makedirs(self._data_dir, exist_ok=True)
        try:
            if self._data_path is not None:
                self._data = csv.read_file(
                    self._data_path,
                    index_column=self.index_column,
                    index_type=self.index_type,
                    timezone=self.timezone,
                    separator=self.separator,
                    decimal=self.decimal,
                    rename=self._build_columns(resources),
                )
        except IOError as e:
            raise ConnectionError(self, str(e))

    def disconnect(self) -> None:
        self._data = None

    def is_connected(self) -> bool:
        return True

    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        def _infer_dates(s=start, e=end) -> Tuple[pd.Timestamp, pd.Timestamp]:
            if all(pd.isna(d) for d in [s, e]):
                n = pd.Timestamp.now(tz=self.timezone)
                s = floor_date(n, timezone=self.timezone, freq=self.freq)
                e = ceil_date(n, timezone=self.timezone, freq=self.freq)
            return s, e

        try:
            if self._data is not None:
                data = self._data
            else:
                data = csv.read_files(
                    self._data_dir,
                    self.freq,
                    self.format,
                    *_infer_dates(),
                    index_column=self.index_column,
                    index_type=self.index_type,
                    timezone=self.timezone,
                    separator=self.separator,
                    decimal=self.decimal,
                )

            if self.index_type in ["timestamp", "unix"] and all(pd.isna(d) for d in [start, end]):
                now = pd.Timestamp.now(tz=self.timezone)
                index = data.index.tz_convert(self.timezone).get_indexer([now], method="nearest")
                data = data.iloc[[index[-1]], :]

            columns = self._build_columns(resources)
            results = []
            for resource in resources:
                resource_column = columns[resource.id]
                if resource_column not in data.columns:
                    results.append(pd.Series(name=resource.id))
                    continue
                resource_data = data.loc[:, resource_column].copy()
                resource_data.name = resource.id
                results.append(resource_data)
            return pd.concat(results, axis="columns")

        except IOError as e:
            raise ConnectionError(self, str(e))

    # noinspection PyTypeChecker
    def read_first(self, resources: Resources) -> Optional[pd.DataFrame]:
        try:
            if self._data is not None:
                data = self._data
            else:
                files = csv.get_files(
                    self._data_dir,
                    self.freq,
                    self.format,
                    timezone=self.timezone,
                )
                if len(files) == 0:
                    return None
                data = csv.read_file(
                    files[0],
                    index_column=self.index_column,
                    index_type=self.index_type,
                    timezone=self.timezone,
                    separator=self.separator,
                    decimal=self.decimal,
                    rename=self._build_columns(resources),
                )
            if data is None or data.empty:
                return None

            columns = self._build_columns(resources)
            results = []
            for resource in resources:
                resource_column = columns[resource.id]
                if resource_column not in data.columns:
                    results.append(pd.Series(name=resource.id))
                    continue
                resource_data = data.loc[:, resource_column].copy()
                resource_data.name = resource.id
                results.append(resource_data)
            return pd.concat(results, axis="columns").head(1)

        except IOError as e:
            raise ConnectionError(self, str(e))

    # noinspection PyTypeChecker
    def read_last(self, resources: Resources) -> Optional[pd.DataFrame]:
        try:
            if self._data is not None:
                data = self._data
            else:
                files = csv.get_files(
                    self._data_dir,
                    self.freq,
                    self.format,
                    timezone=self.timezone,
                )
                if len(files) == 0:
                    return None
                data = csv.read_file(
                    files[-1],
                    index_column=self.index_column,
                    index_type=self.index_type,
                    timezone=self.timezone,
                    separator=self.separator,
                    decimal=self.decimal,
                    rename=self._build_columns(resources),
                )
            if data is None or data.empty:
                return None

            columns = self._build_columns(resources)
            results = []
            for resource in resources:
                resource_column = columns[resource.id]
                if resource_column not in data.columns:
                    results.append(pd.Series(name=resource.id))
                    continue
                resource_data = data.loc[:, resource_column].copy()
                resource_data.name = resource.id
                results.append(resource_data)
            return pd.concat(results, axis="columns").tail(1)

        except IOError as e:
            raise ConnectionError(self, str(e))

    def write(self, data: pd.DataFrame) -> None:
        columns = self._build_columns(self.resources)
        kwargs = {
            "timezone": self.timezone,
            "separator": self.separator,
            "decimal": self.decimal,
            "override": self.override,
            "rename": columns,
        }
        if self.pretty:
            data.index.name = self.index_column.title()
        else:
            data.index.name = self.index_column

        if self.slice:
            csv.write_files(data, self._data_dir, self.freq, self.format, **kwargs)
        else:
            csv_file = os.path.join(self._data_dir, data.index[0].strftime(self.format) + ".csv")
            csv.write_file(data, csv_file, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.simulation.progress
~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import json
import logging
import multiprocessing as process
import os
from typing import Any, Optional


class Progress:
    # noinspection PyUnresolvedReferences
    def __init__(
        self,
        desc: Optional[str] = None,
        total: Optional[int] = None,
        value: Optional[process.Value | int] = None,
        file: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._logger = logging.getLogger(self.__module__)

        # Manually disable the tqdm progress bars for testing on broken windows consoles
        # kwargs['disable'] = True
        # kwargs['file'] = sys.stdout

        if value is not None and type(value).__name__ == "Synchronized" and value.value > 0:
            kwargs["initial"] = value.value

        self._total = total
        self._value = value
        self._file = file
        if os.path.exists(file):
            with open(self._file, "r", encoding="utf-8") as file:
                self.status = json.load(file).get("status", "loaded")
        else:
            self.status = "initialized"
        try:
            from tqdm import tqdm

            self._bar = tqdm(desc=desc, total=total, **kwargs)

        except ImportError as e:
            self._logger.debug("Unable to import tqdm progress library: %s", e)
            self._bar = None

    def close(self) -> None:
        if self._bar:
            self._bar.close()

    # noinspection PyTypeChecker
    def complete(self, status: str = "success", **results: Any) -> None:
        self._update(self._total, status, dump=False)
        if self._file is not None:
            with open(self._file, "w", encoding="utf-8") as file:
                json.dump({"status": status, **results}, file, ensure_ascii=False, indent=4)

    # noinspection PyUnresolvedReferences
    def update(self) -> None:
        if self._value is not None:
            if type(self._value).__name__ == "Synchronized":
                with self._value.get_lock():
                    self._value.value += 1
                    self._update(self._value.value)
            else:
                self._value += 1
                self._update(self._value)
        elif self._bar:
            self._bar.update()

    # noinspection PyTypeChecker
    def _update(self, value: int, status: str = "running", dump: bool = True) -> None:
        progress = value / self._total * 100
        if progress % 1 <= 1 / self._total * 100 and self._file is not None and dump:
            with open(self._file, "w", encoding="utf-8") as file:
                results = {
                    "status": status,
                    "progress": int(progress),
                }
                json.dump(results, file, ensure_ascii=False, indent=4)

        self.status = status
        if self._bar:
            self._bar.update(value - self._bar.n)



# -*- coding: utf-8 -*-
"""
lories._core._context
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Callable, MutableMapping
from itertools import chain
from typing import Any, Collection, Generic, Iterable, Iterator, Optional, Sequence, Tuple, TypeVar, overload

import pandas as pd
from lories._core._entity import Entity, _Entity


# noinspection PyAbstractClass
class _Context(ABC, Generic[Entity], MutableMapping[str, Entity]):
    __map: OrderedDict[str, Entity]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__map = OrderedDict()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(str(c.id) for c in self.__map.values())})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{i} = {repr(c)}" for i, c in self.__map.items())

    def __iter__(self) -> Iterator[str]:
        return iter(self.__map.keys())

    def __len__(self) -> int:
        return len(self.__map)

    def __contains__(self, __object: str | Entity) -> bool:
        return self._contains(__object)

    def __getitem__(self, __uid: Iterable[str] | str) -> Entity | Collection[Entity]:
        if isinstance(__uid, str):
            return self._get(__uid)
        if isinstance(__uid, Iterable):
            return [self._get(i) for i in __uid]
        raise KeyError(__uid)

    def __setitem__(self, __uid: str, __object: Entity) -> None:
        self._set(__uid, __object)

    def __delitem__(self, __uid: str) -> None:
        self._remove(__uid)

    def _contains(self, __object: str | Entity) -> bool:
        if isinstance(__object, str):
            return __object in self.__map.keys()
        if isinstance(__object, _Entity):
            return __object in self.__map.values()
        return False

    def _get(self, __uid: str) -> Entity:
        return self.__map[__uid]

    def _set(self, __uid: str, __object: Entity) -> None:
        if __uid in self.keys():
            raise ValueError(f'Entity with ID "{__uid}" already exists')

        self.__map[__uid] = __object

    def _add(self, *__objects: Entity) -> None:
        for __object in __objects:
            self._set(str(__object.id), __object)

    @abstractmethod
    def _create(self, *args, **kwargs) -> Entity: ...

    @abstractmethod
    def _update(self, *args, **kwargs) -> None: ...

    def _remove(self, *__objects: str | Entity) -> None:
        for __object in __objects:
            if isinstance(__object, str):
                del self.__map[__object]
            elif isinstance(__object, _Entity):
                del self.__map[__object.id]

    def sort(self):
        def order(text: str) -> Tuple[Any, ...]:
            elements = re.split(r"[^0-9A-Za-zäöüÄÖÜß]+", text)
            elements = list(chain(*[re.findall(r"\D+|\d+", t) for t in elements]))
            elements = [(int(t), 0) if t.isdigit() else (0, t) for t in elements if pd.notna(t) and t.strip()]
            return tuple(elements)

        self.__map = OrderedDict(sorted(self.__map.items(), key=lambda e: order(e[0])))

    # noinspection PyShadowingBuiltins
    @overload
    def filter(self, filter: Optional[Callable[[Entity], bool]]) -> Sequence[Entity]: ...

    def filter(self, *filters: Optional[Callable[[Entity], bool]]) -> Sequence[Entity]:
        def _filters(__object: Entity) -> bool:
            for _filter in filters:
                if _filter is not None and not _filter(__object):
                    return False
            return True

        return [c for c in self.__map.values() if _filters(c)]


Context = TypeVar("Context", bound=_Context)



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.layout
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from .card import (  # noqa: F401
    PageCard,
    PageCardItem,
)
from .page import PageLayout  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.components.tariff
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from . import tariff  # noqa: F401
from .tariff import (  # noqa: F401
    Tariff,
    register_tariff_type,
    registry,
)

from . import provider  # noqa: F401
from .provider import TariffProvider  # noqa: F401

import importlib

for import_provider in ["static", "entsoe"]:
    try:
        importlib.import_module(f".{import_provider}", "lories.components.tariff")

    except ModuleNotFoundError:
        # TODO: Implement meaningful logging here
        pass

del importlib



# -*- coding: utf-8 -*-
"""
lories.connectors.errors
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories._core._connector import Connector  # noqa
from lories.core import ResourceError, ResourceUnavailableError


class ConnectorError(ResourceError):
    """
    Raise if an error occurred accessing the connector.

    """

    # noinspection PyArgumentList
    def __init__(self, connector: Connector, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.connector = connector


class ConnectorUnavailableError(ResourceUnavailableError, ConnectorError):
    """
    Raise if an accessed connector can not be found.

    """


# noinspection PyShadowingBuiltins
class ConnectionError(ConnectorError, ConnectionError):
    """
    Raise if an error occurred with the connection.

    """


class DatabaseException(ConnectorError):
    """
    Raise if an error occurred accessing the database.

    """


class DatabaseUnavailableException(ConnectorUnavailableError):
    """
    Raise if an accessed database can not be found.

    """



# -*- coding: utf-8 -*-
"""
lories.data.converters.io.current
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories.core import ConfigurationError
from lories.core.typing import Configurations
from lories.data.converters import register_converter_type
from lories.data.converters.io.analog import AnalogInput


# noinspection PyAbstractClass
@register_converter_type("current_input", "current_sensor")
class CurrentInput(AnalogInput):
    INPUT_KEY: str = "current"


# noinspection PyAbstractClass
@register_converter_type("current_loop", "4_20ma_input")
class CurrentLoopInput(AnalogInput):
    def _configure_input(
        self,
        configs: Configurations,
        default_max: float = 0.02,
        default_min: float = 0.004,
    ) -> None:
        super()._configure_input(configs, default_max, default_min)

    # noinspection PyShadowingBuiltins
    def _assert_input(self, min: float, max: float, zero: float) -> None:
        super()._assert_input(min, max, zero)
        if min < 0.004:
            raise ConfigurationError(f"Invalid current minimum for '{self.id}' 4-20mA input: {min}")
        if max > 0.02:
            raise ConfigurationError(f"Invalid current maximum for '{self.id}' 4-20mA input: {max}")



# -*- coding: utf-8 -*-
"""
lories.connectors.access
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories._core._component import Component  # noqa
from lories._core._connector import Connector, _Connector, _ConnectorContext  # noqa
from lories._core._data import _DataManager  # noqa
from lories.core import RegistratorAccess, ResourceError
from lories.util import get_context


class ConnectorAccess(_ConnectorContext, RegistratorAccess[Connector]):
    # noinspection PyUnresolvedReferences
    def __init__(self, registrar: Component, **kwargs) -> None:
        context = get_context(registrar, _DataManager).connectors
        super().__init__(context, registrar, **kwargs)

    # noinspection PyProtectedMember, PyShadowingBuiltins
    def _set(self, id: str, connector: Connector) -> None:
        if not isinstance(connector, _Connector):
            raise ResourceError(f"Invalid connector type: {type(connector)}")

        super()._set(id, connector)



# -*- coding: utf-8 -*-
"""
lories.io.plot
~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
import os
from typing import List, Literal, Optional, Tuple

import matplotlib.pyplot as plt
import seaborn as sns

import numpy as np
import pandas as pd

logging.getLogger("PIL.PngImagePlugin").setLevel(logging.WARN)
logging.getLogger("matplotlib").setLevel(logging.WARN)
logger = logging.getLogger(__name__)

COLORS = [
    "#004F9E",
    "#FFB800",
]

INCH = 2.54
WIDTH = 32
HEIGHT = 9


# noinspection PyDefaultArgument, SpellCheckingInspection
def line(
    x: Optional[pd.Series | str] = None,
    y: Optional[pd.DataFrame | pd.Series | str] = None,
    data: Optional[pd.DataFrame] = None,
    title: str = "",
    xlabel: str = "",
    ylabel: str = "",
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    grids: Optional[Literal["both", "x", "y"]] = None,
    colors: List[str] = COLORS,
    palette: Optional[str] = None,
    hue: Optional[str] = None,
    width: int = WIDTH,
    height: int = HEIGHT,
    show: bool = False,
    file: str = None,
    **kwargs,
) -> None:
    plt.figure(figsize=(width / INCH, height / INCH), dpi=120, tight_layout=True)

    color_num = max(len(np.unique(data[hue])) if hue else len(data.columns) - 1, 1)
    if color_num > 1:
        if palette is None:
            palette = f"blend:{','.join(colors)}"
        kwargs["palette"] = sns.color_palette(palette, n_colors=color_num)
        kwargs["hue"] = hue
    else:
        kwargs["color"] = colors[0]

    plot = sns.lineplot(
        x=x,
        y=y,
        data=data,
        errorbar=("pi", 50),
        estimator=np.median,
        **kwargs,
    )

    if isinstance(x, str) and x in ["hour", "horizon"]:
        index_unique = data[x].astype(int).unique()
        index_unique.sort()
        plt.xticks(index_unique, labels=index_unique)

    plot.set(xlabel=xlabel, ylabel=ylabel, title=title)
    plt.box(on=False)

    if xlim is not None:
        plt.xlim(xlim)
    if ylim is not None:
        plt.ylim(ylim)

    if grids is not None:
        plt.grid(color="grey", linestyle="--", linewidth=0.25, alpha=0.5, axis=grids)

    if show:
        plt.show()
    if file is not None:
        plot.figure.savefig(file)

    plt.close(plot.figure)
    plt.clf()


# noinspection PyDefaultArgument, SpellCheckingInspection
def bar(
    x: Optional[pd.Index | pd.Series | str] = None,
    y: Optional[pd.DataFrame | pd.Series | str] = None,
    data: Optional[pd.DataFrame] = None,
    title: str = "",
    xlabel: str = "",
    ylabel: str = "",
    label_type: Optional[Literal["edge", "center"]] = None,
    colors: List[str] = COLORS,
    palette: Optional[str] = None,
    hue: Optional[str] = None,
    width: int = WIDTH,
    height: int = HEIGHT,
    show: bool = False,
    file: str = None,
    **kwargs,
) -> None:
    plt.figure(figsize=(width / INCH, height / INCH), dpi=120, tight_layout=True)

    color_num = max(len(np.unique(data[hue])) if hue else len(data.columns) - 1, 1)
    if color_num > 1:
        if palette is None:
            palette = f"blend:{','.join(colors)}"
        kwargs["palette"] = sns.color_palette(palette, n_colors=color_num)
        kwargs["hue"] = hue
    else:
        kwargs["color"] = colors[0]

    plot = sns.barplot(x=x, y=y, data=data, **kwargs)

    if (isinstance(x, str) and len(data[x]) > 24) or (
        isinstance(x, (pd.Index, pd.Series, np.ndarray)) and len(np.unique(x)) > 24
    ):
        plot.xaxis.set_tick_params(rotation=45)

    plot.set(xlabel=xlabel, ylabel=ylabel, title=title)

    plt.box(on=False)

    if label_type is not None:
        plot.bar_label(plot.containers[0], label_type=label_type)

    if show:
        plt.show()
    if file is not None:
        plot.figure.savefig(file)

    plt.close(plot.figure)
    plt.clf()


# noinspection PyDefaultArgument, SpellCheckingInspection
def quartiles(
    x: Optional[pd.Series | str] = None,
    y: Optional[pd.DataFrame | pd.Series | str] = None,
    data: Optional[pd.DataFrame] = None,
    title: str = "",
    xlabel: str = "",
    ylabel: str = "",
    method: str = "bar",
    colors: List[str] = COLORS,
    palette: Optional[str] = None,
    hue: Optional[str] = None,
    width: int = WIDTH,
    height: int = HEIGHT,
    show: bool = False,
    file: str = None,
    **kwargs,
) -> None:
    plt.figure(figsize=(width / INCH, height / INCH), dpi=120, tight_layout=True)

    color_num = max(len(np.unique(data[hue])) if hue else len(data.columns) - 1, 1)
    if color_num > 1:
        if palette is None:
            palette = f"blend:{','.join(colors)}"
        kwargs["palette"] = sns.color_palette(palette, n_colors=color_num)
        kwargs["hue"] = hue
    else:
        kwargs["color"] = colors[0]

    if method in ["bar", "bars"]:
        fliers = dict(marker="o", markersize=3, markerfacecolor="none", markeredgecolor="lightgrey")
        plot = sns.boxplot(
            x=x,
            y=y,
            data=data,
            flierprops=fliers,
            **kwargs,
        )

        if (isinstance(x, str) and len(data[x]) > 24) or (isinstance(x, pd.Series) and len(np.unique(x)) > 24):
            plot.xaxis.set_tick_params(rotation=45)

    elif method == "line":
        # stats = data.groupby([x]).describe()
        # index_values = stats.index
        # index_unique = index_values.astype(int).unique().values
        # index_unique.sort()
        #
        # medians = stats[(y, '50%')]
        # quartile1 = stats[(y, '25%')]
        # quartile3 = stats[(y, '75%')]

        plot = sns.lineplot(
            x=x,
            y=y,
            data=data,
            errorbar=("pi", 50),
            estimator=np.median,
            **kwargs,
        )

        # plot.fill_between(index_values, quartile1, quartile3, color=color_palette[0], alpha=0.3)

        if isinstance(x, str) and x in ["hour", "horizon"]:
            index_unique = data[x].astype(int).unique()
            index_unique.sort()
            plot.set_xticks(index_unique, labels=index_unique)

    else:
        logger.error(f'Invalid boxplot method "{method}"')
        return

    plot.set(xlabel=xlabel, ylabel=ylabel, title=title)

    if show:
        plt.show()
    if file is not None:
        plot.figure.savefig(file)

    plt.close(plot.figure)
    plt.clf()


def histograms(data: pd.DataFrame, bins: int = 100, show: bool = False, path: str = "") -> None:
    for column in data.columns:
        plt.figure(figsize=(WIDTH, HEIGHT), dpi=120, tight_layout=True)

        # Create equal space bin values per column
        bin_data = []
        bin_domain = data[column].max() - data[column].min()
        bin_step = bin_domain / bins

        counter = data[column].min()
        for i in range(bins):
            bin_data.append(counter)
            counter = counter + bin_step

        # Add the last value of the counter
        bin_data.append(counter)

        bin_values, bin_data, patches = plt.hist(data[column], bins=bin_data)
        count_range = max(bin_values) - min(bin_values)
        sorted_values = list(bin_values)
        sorted_values.sort(reverse=True)

        # Scale plots by stepping through sorted bin_data
        for i in range(len(sorted_values) - 1):
            if abs(sorted_values[i] - sorted_values[i + 1]) / count_range < 0.80:
                continue
            else:
                plt.ylim([0, sorted_values[i + 1] + 10])
                break

        # Save histogram to appropriate folder
        path_dist = os.path.join(path, "dist")
        path_file = os.path.join(path_dist, "{}.png".format(column))
        if not os.path.isdir(path_dist):
            os.makedirs(path_dist, exist_ok=True)

        plt.title(r"Histogram of " + column)
        plt.savefig(path_file)

        if show:
            plt.show()
        plt.close()
        plt.clf()



# -*- coding: utf-8 -*-
"""
lories.connectors.serial._serial
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import AnyStr, Optional

import serial
from serial import SerialException

from lories.connectors import ConnectionError, Connector
from lories.core import Configurations, Resources


# noinspection PyAbstractClass, SpellCheckingInspection
class _SerialConnector(Connector):
    _serial: Optional[serial.Serial]

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._serial = serial.Serial(
            baudrate=configs.get_int("baudrate", default=9600),
            bytesize=configs.get_int("bytesize", default=serial.EIGHTBITS),
            parity=configs.get("parity", default=serial.PARITY_NONE),
            stopbits=configs.get_int("stopbits", default=serial.STOPBITS_ONE),
            timeout=configs.get_float("timeout", default=3),
            xonxoff=configs.get_bool("xonxoff", default=False),
            rtscts=configs.get_bool("rtscts", default=False),
            dsrdtr=configs.get_bool("dsrdtr", default=False),
        )
        self._serial.port = configs.get("port")

    def connect(self, resources: Resources) -> None:
        try:
            self._serial.open()

        except SerialException as e:
            raise ConnectionError(f"Failed to open serial port {self._serial.port}: {e}") from e

    def disconnect(self) -> None:
        if self.is_connected():
            self._serial.close()

    def is_connected(self) -> bool:
        return self._serial is not None and self._serial.is_open

    def _write_string(self, data: AnyStr, encode="ascii") -> None:
        self._serial.write(data.encode(encode))

    def _read_line(self, decode="ascii") -> AnyStr:
        return self._serial.readline().decode(decode, errors="ignore").replace('\x00', '').strip()



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.components.registry
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Collection, Dict, Generic, List, Optional, Type, TypeVar

from lories.application.view.pages import Page, PageGroup
from lories.core.errors import ResourceError
from lories.util import validate_key

P = TypeVar("P", bound=Page)
G = TypeVar("G", bound=PageGroup)


# noinspection PyShadowingBuiltins
class _PageRegistration(ABC, Generic[P]):
    _class: Type[P]
    _factory: callable

    _kwargs: Dict[str, Any]

    def __init__(
        self,
        cls: Type[P],
        factory: Optional[Callable] = None,
        **kwargs,
    ):
        self._class = cls
        self._factory = cls if factory is None else factory
        self._kwargs = kwargs

    @abstractmethod
    def has_type(self, *args) -> bool:
        pass

    def initialize(self, *args, **kwargs) -> P:
        factory = self._factory
        if factory is None:
            factory = self._class
        elif not callable(factory):
            raise ResourceError(f"Invalid registration initialization function: {factory}")

        return factory(*args, **kwargs, **self._kwargs)


# noinspection PyShadowingBuiltins
class PageRegistration(_PageRegistration[P]):
    type: Type

    def __init__(
        self,
        cls: Type[P],
        type: Type,
        factory: Optional[Callable] = None,
        **kwargs,
    ):
        super().__init__(cls, factory, **kwargs)
        self.type = type

    def has_type(self, type: Type) -> bool:
        return issubclass(type, self.type)


# noinspection PyShadowingBuiltins
class GroupRegistration(_PageRegistration[G]):
    types: List[Type]

    key: str
    name: str

    def __init__(
        self,
        cls: Type[G],
        *types: Type,
        key: Optional[str] = None,
        name: Optional[str] = None,
        factory: Optional[Callable] = None,
    ):
        super().__init__(cls, factory)
        self.types = list(types)

        if name is None:
            if len(types) > 1:
                raise ValueError(f"Ambiguous ID for {len(types)} types of class: {cls.__init__()}")
            name = types[0].__name__
        self.name = name

        if key is None:
            key = validate_key(name)
        self.key = key

    def add_type(self, type: Type) -> None:
        self.types.append(type)

    def has_type(self, *types: Type) -> bool:
        _types = [t for t in self.types if any(issubclass(_t, t) for _t in types)]
        return len(_types) > 0


# noinspection PyShadowingBuiltins
class PageRegistry:
    pages: Collection[PageRegistration]
    groups: Collection[GroupRegistration]

    def __init__(self) -> None:
        self.pages: List[PageRegistration] = []
        self.groups: List[GroupRegistration] = []

    # noinspection PyTypeChecker, PyProtectedMember
    def register_page(
        self,
        cls: Type[G],
        type: Type,
        factory: Optional[Callable] = None,
        replace: bool = False,
    ) -> None:
        if not issubclass(cls, Page):
            raise ValueError("Can only register Page types")
        existing = self._get_pages(type)
        if len(existing) > 0:
            if replace:
                for page in existing:
                    self.pages.remove(page)
            else:
                raise ResourceError(
                    f"Registration for '{type}' does already exist: " ", ".join(p._class.__name__ for p in existing)
                )
        self.pages.append(PageRegistration(cls, type, factory=factory))

    # noinspection PyTypeChecker, PyProtectedMember, PyUnresolvedReferences
    def register_group(
        self,
        cls: Type[G],
        *types: Type,
        key: Optional[str] = None,
        name: Optional[str] = None,
        factory: Optional[Callable] = None,
        replace: bool = False,
    ) -> None:
        if not issubclass(cls, PageGroup):
            raise ValueError("Can only register PageGroup types")
        existing = self._get_groups(*types)
        if len(existing) > 0:
            if replace:
                for group in existing:
                    self.groups.remove(group)
            else:
                raise ResourceError(
                    f"Registration for types "
                    f"'{', '.join(t.__name__ for t in types)}' does already exist: "
                    ", ".join(p._class.__name__ for p in existing)
                )
        self.groups.append(GroupRegistration(cls, *types, key=key, name=name, factory=factory))

    def has_page(self, *types: Type) -> bool:
        return len(self._get_pages(*types)) > 0

    def _get_pages(self, *types: Type) -> Collection[PageRegistration]:
        return [p for p in self.pages if p.has_type(*types)]

    def get_page(self, type: Type) -> PageRegistration:
        for page in self.pages:
            if page.has_type(type):
                return page
        raise ValueError(f"Registration '{type}' does not exist")

    def has_group(self, *types: Type) -> bool:
        return len(self._get_groups(*types)) > 0

    def _get_groups(self, *types: Type) -> Collection[GroupRegistration]:
        return [g for g in self.groups if g.has_type(*types)]

    def get_group(self, type: Type) -> GroupRegistration:
        for group in self.groups:
            if group.has_type(type):
                return group
        raise ValueError(f"Registration '{type}' does not exist")



# -*- coding: utf-8 -*-
"""
lories.application.view.pages
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from . import layout  # noqa: F401
from .layout import (  # noqa: F401
    PageCard,
    PageLayout,
)

from . import page  # noqa: F401
from .page import (  # noqa: F401
    Page,
    PageException,
)

from . import group  # noqa: F401
from .group import PageGroup  # noqa: F401

from . import components  # noqa: F401
from .components import (  # noqa: F401
    ComponentPage,
    ComponentGroup,
)

from .header import PageHeader  # noqa: F401
from .footer import PageFooter  # noqa: F401

from .registry import PageRegistry  # noqa: F401

from .view import (  # noqa: F401
    View,
    register_component_page,
    register_component_group,
    registry,
)

from .components import system  # noqa: F401
from .components.system import SystemPage  # noqa: F401

from .components import weather  # noqa: F401
from .components.weather import WeatherPage  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories._core._activator
~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from typing import TypeVar

from lories._core._configurator import _Configurator


class _Activator(_Configurator):
    @abstractmethod
    def is_active(self) -> bool: ...

    def activate(self) -> None:
        pass

    def deactivate(self) -> None:
        pass


Activator = TypeVar("Activator", bound=_Activator)



# -*- coding: utf-8 -*-
"""
lories.settings
~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging.config
import os
import shutil
import sys
from argparse import ArgumentParser
from typing import Any, Dict

from lories.core import Configurations, Directories, Directory
from lories.util import to_bool


class Settings(Configurations):
    app: str

    # noinspection PyProtectedMember, SpellCheckingInspection
    def __init__(self, app_name: str, app_file: str = "settings.conf", parser: ArgumentParser = None) -> None:
        app_args = _parse_kwargs(parser)
        app_args["name"] = app_name
        app_args["systems"] = {k: to_bool(app_args.pop(f"system_{k}", False)) for k in ["scan", "flat", "copy"]}
        app_dirs = Directories(**{d: app_args.pop(d, None) for d in Directories.KEYS})
        super().__init__(app_file, app_dirs, app_args)
        self._load(require=False)
        self._load_logging()

        override_path = os.path.join(self.dirs.data, self.name)
        if os.path.isfile(override_path):
            self._load_toml(override_path)
            self.dirs.update(self.get_member(Directories.TYPE, defaults={}))
        if self.dirs.conf._dir is None:
            self.dirs._conf = Directory(os.path.dirname(self.path), default="conf")

    def __getattr__(self, attr):
        # __getattr__ gets called when the item is not found via __getattribute__
        # To avoid recursion, call __getattribute__ directly to get components dict
        configs = Configurations.__getattribute__(self, f"_{Configurations.__name__}__configs")
        if attr in configs.keys():
            return configs[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no configuration '{attr}'")

    @property
    def _members_dir(self) -> Directory:
        return self.dirs.conf

    def _load_logging(self) -> None:
        logging_file = os.path.join(self.dirs.conf, "logging.conf")
        if not os.path.isfile(logging_file) and not self.dirs.conf.is_default():
            logging_default = logging_file.replace("logging.conf", "logging.default.conf")
            if os.path.isfile(logging_default):
                shutil.copy(logging_default, logging_file)

        if os.path.isfile(logging_file):
            logging.config.fileConfig(logging_file)
            for handler in logging.getLoggerClass().root.handlers:
                if isinstance(handler, logging.FileHandler):
                    self.dirs.log = os.path.dirname(handler.baseFilename)
                    break
            if not os.path.isdir(self.dirs.log):
                os.makedirs(self.dirs.log, exist_ok=True)
        else:
            handler_console = logging.StreamHandler(sys.stdout)
            handler_console.setLevel(logging.INFO)
            handler_console.setFormatter(
                logging.Formatter("%(asctime)s.%(msecs)03d - %(name)s - %(message)s", "%Y-%m-%d %H:%M:%S")
            )

            # TODO: Think about fallback logging configuration necessity to use default logfile location
            # log_file = app_name.lower()
            # if not log_file.endswith('.log'):
            #     log_file += '.log'
            # handler_file = logging.FileHandler(os.path.join(self.dirs._log or 'log', log_file))
            # handler_file.setLevel(logging.WARN)
            # handler_file.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(message)s",
            #                                             "%Y-%m-%d %H:%M:%S"))
            logging.basicConfig(
                force=True,
                level=logging.INFO,
                handlers=[
                    handler_console,
                    # handler_file,
                ],
            )


# noinspection PyProtectedMember
def _parse_kwargs(parser: ArgumentParser) -> Dict[str, Any]:
    if parser is not None:
        action_parsers = [a for a in parser._actions if a.dest == "action"]
        if len(action_parsers) > 0:
            action_parser = next(iter(action_parsers))
        else:
            action_parser = parser.add_subparsers(dest="action")
        if not action_parser.required and action_parser.default is None:
            action_parser.required = True
        if "run" not in action_parser.choices.keys():
            action_parser.add_parser("run", help="run local resources, connectors and systems")
        if "start" not in action_parser.choices.keys():
            action_parser.add_parser("start", help="start the local resource system")
        if "rotate" not in action_parser.choices.keys():
            replicate_parser = action_parser.add_parser(
                name="rotate",
                help="rotate data of a local database",
            )
            replicate_parser.add_argument(
                "--full",
                dest="full",
                action="store_true",
                help="flags if the rotation should be executed for all data",
            )
        if "replicate" not in action_parser.choices.keys():
            replicate_parser = action_parser.add_parser(
                name="replicate",
                help="replicate data from a remote database",
            )
            replicate_parser.add_argument(
                "--force",
                dest="force",
                action="store_true",
                help="flags if the replication should force rewrite data for checksum mismatches",
            )
            replicate_parser.add_argument(
                "--full",
                dest="full",
                action="store_true",
                help="flags if the replication should be executed for all data",
            )
        if "simulate" not in action_parser.choices.keys():
            simulate_parser = action_parser.add_parser(
                name="simulate",
                help="simulate local resources and systems for a specified time range",
            )
            simulate_parser.add_argument(
                "--start",
                dest="start",
                metavar="datetime",
                help="the start datetime for the simulation to start",
            )
            simulate_parser.add_argument(
                "--end",
                dest="end",
                metavar="datetime",
                help="the end datetime for the simulation to end",
            )

        parser.add_argument(
            "-c",
            "--conf-dir",
            dest="conf_dir",
            metavar="dir",
            help="directory to expect root configuration files",
        )
        parser.add_argument(
            "-d",
            "--data-dir",
            dest="data_dir",
            metavar="dir",
            help="directory to expect and write data files to",
        )
        parser.add_argument(
            "--system-scan",
            dest="system_scan",
            action="store_true",
            help="flags whether several systems will be expected, instead of a single one",
        )
        parser.add_argument(
            "--system-flat",
            dest="system_flat",
            action="store_true",
            help="flags if the configuration files will be expected directly in the data directory, "
            + "instead of a corresponding 'conf' directory",
        )
        parser.add_argument(
            "--system-copy",
            dest="system_copy",
            action="store_true",
            help="flags if the configured system files should be copied to the specified data directory if it is empty",
        )

        args = parser.parse_args()
        kwargs = vars(args)
    else:
        kwargs = {}

    return kwargs



# -*- coding: utf-8 -*-
"""
lories.connectors.tasks.read
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import inspect
from typing import Optional

import pandas as pd
from lories._core._channel import ChannelState  # noqa
from lories.connectors.tasks.task import ConnectorTask


class ReadTask(ConnectorTask):
    # noinspection PyArgumentList
    def run(self, inplace: bool = False, **kwargs) -> Optional[pd.DataFrame]:
        self._logger.debug(
            f"Reading {len(self.channels)} channels of '{type(self.connector).__name__}': {self.connector.id}"
        )
        signature = inspect.signature(type(self.connector).read)
        arguments = [p.name for p in signature.parameters.values() if p.kind == p.POSITIONAL_OR_KEYWORD]
        for argument in list(kwargs.keys()):
            if argument not in arguments:
                value = kwargs.pop(argument)
                self._logger.warning(
                    f"Trying to read Connector '{self.connector.id}' with unknown argument '{argument}': {value}"
                )

        data = self.connector.read(self.channels, **kwargs)
        if data is None or data.dropna(axis="columns", how="all").empty:
            if inplace:
                self.channels.set_state(ChannelState.NOT_AVAILABLE)
            return None
        if inplace:
            self.channels.set_frame(data)
        return data



# -*- coding: utf-8 -*-
"""
lories.components.context
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from typing import Callable, Optional, Type

from lories._core._component import Component, _Component, _ComponentContext  # noqa
from lories.core.configs import Configurations
from lories.core.register import RegistratorContext, Registry

registry = Registry[_Component]()


# noinspection PyShadowingBuiltins
def register_component_type(
    type: str,
    *alias: str,
    factory: Callable[[ComponentContext | Component, Optional[Configurations]], Component] = None,
    replace: bool = False,
) -> Callable[[Type[Component]], Type[Component]]:
    # noinspection PyShadowingNames
    def _register(cls: Type[Component]) -> Type[Component]:
        registry.register(cls, type, *alias, factory=factory, replace=replace)
        return cls

    return _register


class ComponentContext(_ComponentContext, RegistratorContext[Component]):
    @property
    def _registry(self) -> Registry[Component]:
        return registry

    # noinspection PyShadowingBuiltins
    def activate(self, filter: Optional[Callable[[Component], bool]] = None) -> None:
        self._activate(*self.filter(filter))

    def _activate(self, *component: Component) -> None:
        for component in component:
            if not component.is_enabled():
                self._logger.debug(
                    f"Skipping to activate disabled {type(component).__name__} '{component.name}': {component.id}"
                )
                continue
            self.__activate(component)

    def __activate(self, component: Component) -> None:
        self._logger.debug(f"Activating {type(component).__name__} '{component.name}': {component.id}")
        component.activate()

        self._logger.info(f"Activated {type(component).__name__} '{component.name}': {component.id}")

    # noinspection PyShadowingBuiltins
    def deactivate(self, filter: Optional[Callable[[Component], bool]] = None) -> None:
        self._deactivate(*self.filter(filter))

    def _deactivate(self, *components: Component) -> None:
        for component in reversed(list(components)):
            if not component.is_active():
                self._logger.debug(
                    f"Skipping to deactivate already deactivated {type(component).__name__} '{component.name}': "
                    f"{component.id}"
                )
                return
            self.__deactivate(component)

    def __deactivate(self, component: Component) -> None:
        if not component.is_active():
            self._logger.debug(
                f"Skipping to deactivate already deactivated {type(component).__name__} '{component.name}': "
                f"{component.id}"
            )
            return
        try:
            self._logger.debug(f"Deactivating {type(component).__name__} '{component.name}': {component.id}")
            component.deactivate()

            self._logger.info(f"Deactivated {type(component).__name__} '{component.name}': {component.id}")

        except Exception as e:
            self._logger.warning(f"Failed deactivating component '{component.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)



# -*- coding: utf-8 -*-
"""
lories.simulation
~~~~~~~~~~~~~~~~~


"""

from . import progress  # noqa: F401
from .progress import Progress  # noqa: F401

from . import durations  # noqa: F401
from .durations import Durations  # noqa: F401

from . import result  # noqa: F401
from .result import Result  # noqa: F401

from . import results  # noqa: F401
from .results import Results  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.data.converters.errors
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories.core.errors import ResourceError


class ConversionError(ResourceError):
    """
    Raise if a conversion failed

    """



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.components.system
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories.application.view.pages import PageLayout, register_component_page
from lories.application.view.pages.components import ComponentGroup
from lories.system import System


@register_component_page(System)
class SystemPage(ComponentGroup[System]):
    @property
    def path(self) -> str:
        return f"/system/{self._encode_id(self.key)}"

    def create_layout(self, layout: PageLayout) -> None:
        super().create_layout(layout)
        # TODO: Implement location



# -*- coding: utf-8 -*-
"""
lories.connectors.entsoe
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Optional

# TODO: Add to requirements
from entsoe import EntsoePandasClient
from entsoe.mappings import Area as EntsoeArea
from requests.exceptions import HTTPError
from urllib3.exceptions import MaxRetryError

import pandas as pd
import pytz as tz
from lories.connectors import ConnectionError, Connector, ConnectorError
from lories.core.configs import ConfigurationError
from lories.typing import Configurations, Resources, Timestamp
from lories.util import parse_freq

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


# noinspection SpellCheckingInspection
class EntsoeConnector(Connector):
    DAY_AHEAD: str = "day_ahead"
    METHODS: list = [DAY_AHEAD]

    resolution: Literal["60min", "30min", "15min"] = "60min"

    country_code: str
    _api_key: str

    _client: Optional[EntsoePandasClient] = None

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self.resolution = self._validate_resolution(configs.get("resolution", default=EntsoeConnector.resolution))
        self.country_code = self._validate_country_code(configs.get("country_code").upper())
        self._api_key = configs.get("api_key")
        if self._api_key is None:
            raise ConfigurationError("Missing security token")

    # noinspection PyTypeChecker
    def _validate_resolution(self, resolution: str) -> Literal["60min", "30min", "15min"]:
        resolution = parse_freq(resolution)
        if resolution not in ["60min", "30min", "15min"]:
            raise ConnectorError(self, f"Invalid resolution: {resolution}.")
        return resolution

    def _validate_country_code(self, country_code) -> str:
        """Validate the country code against the Entsoe mappings."""
        if country_code is None:
            raise ConfigurationError(self, "Missing country code")
        elif not (EntsoeArea.has_code(country_code) or country_code == "DE"):
            raise ConfigurationError(self, f"Invalid country code: {country_code}.")
        return country_code

    def connect(self, resources: Resources) -> None:
        self._client = EntsoePandasClient(api_key=self._api_key)

    def disconnect(self) -> None:
        if self._client is not None:
            self._client = None

    def is_connected(self) -> bool:
        return self._client is not None

    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        if start is None:
            start = pd.Timestamp.now(tz=tz.UTC).floor(self.resolution)
        if end is None:
            end = start + pd.Timedelta(days=1) - pd.Timedelta(self.resolution)

        results = []
        country_code = self._get_country_code(self.country_code, start, end)
        for method, method_resources in resources.groupby("method"):
            if method not in self.METHODS:
                raise ConnectorError(self, f"Unsupported method: {method}. Supported methods: {self.METHODS}")

            if method.lower() == EntsoeConnector.DAY_AHEAD:
                # TODO: Exception handling
                try:
                    prices = self._client.query_day_ahead_prices(country_code, start=start, end=end)

                except (MaxRetryError, HTTPError) as e:
                    if isinstance(e, MaxRetryError):
                        raise ConnectionError(self, str(e))
                    elif isinstance(e, HTTPError) and "Unauthorized" in str(e):
                        raise ConnectorError(self, "Unauthorized access. Check your API key.")
                    raise ConnectorError(self, str(e))

                result = pd.DataFrame()
                for resource in method_resources:
                    result[resource.id] = prices
                results.append(result)

        if len(results) == 0:
            return pd.DataFrame()

        results = sorted(results, key=lambda d: min(d.index))
        data = pd.concat(results, axis="columns")
        return data

    def write(self, data: pd.DataFrame) -> None:
        raise NotImplementedError("EntsoeConnector does not support writing data")

    def _get_country_code(self, county_code: str, start: Timestamp, end: Timestamp) -> str:
        """
        Returns the country code for Germany based on the start and end dates.
        """
        # https://github.com/EnergieID/entsoe-py?tab=readme-ov-file
        #       	2015	2016	2017	2018	2019	2020	2021
        # DE	    no	    no  	no  	no  	no  	no  	no
        # DE_AT_LU	yes	    yes	    yes	    yes	    no	    no	    no
        # DE_LU	    no	    no  	no  	yes 	yes 	yes 	yes
        # AT	    no	    no  	no  	yes 	yes 	yes 	yes

        if county_code in ["DE", "AT", "LU"]:
            if end.year < 2019:
                return "DE_AT_LU"
            elif start.year > 2018:
                if county_code in ["DE", "LU"]:
                    return "DE_LU"
                else:
                    return "AT"
            elif start.year < 2018 and 2018 < end.year:
                raise ConnectorError(self, "Cannot determine country code for Germany over the given time range.")

            if county_code in ["DE", "LU"]:
                return "DE_LU"
            else:
                return "AT"

        return county_code



# -*- coding: utf-8 -*-
"""
lories.connectors.tasks
~~~~~~~~~~~~~~~~~~~~~~~


"""

from .connect import ConnectTask  # noqa: F401
from .check import CheckTask  # noqa: F401
from .read import ReadTask  # noqa: F401
from .write import WriteTask  # noqa: F401
from .log import LogTask  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.application.main
~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import sys
import traceback
from typing import Optional, Type

import pandas as pd
from lories import ConfigurationUnavailableError, Settings, System
from lories.application import Interface
from lories.connectors import Database, DatabaseException
from lories.data.manager import DataManager
from lories.simulation import Results
from lories.typing import Timestamp
from lories.util import slice_range, to_timedelta


class Application(DataManager):
    _interface: Optional[Interface] = None

    @classmethod
    def load(cls, name: str, factory: Type[System] = System, **kwargs) -> Application:
        settings = Settings(name, **kwargs)
        app = cls(settings)
        app.configure(settings, factory)
        return app

    # noinspection PyProtectedMember
    def __init__(self, settings: Settings, **kwargs) -> None:
        super().__init__(settings, name=settings["name"], **kwargs)
        if not settings.has_member(Interface.TYPE):
            settings._add_member(Interface.TYPE, {"enabled": False})

        # Check if the tasked action may be headless
        if settings.get("action").lower() == "start":
            self._interface = Interface(self, settings.get_member(Interface.TYPE))

    # noinspection PyProtectedMember, PyTypeChecker, PyMethodOverriding
    def configure(self, settings: Settings, factory: Type[System] = System) -> None:
        super().configure(settings)
        self._logger.debug(f"Setting up {type(self).__name__}: {self.name}")
        components = []
        try:
            system_dirs = settings.dirs.to_dict()
            system_dirs["conf_dir"] = None
            systems_configs = settings.get_member("systems")
            systems_flat = systems_configs.get_bool("flat")
            if systems_configs.get_bool("scan"):
                if systems_configs.get_bool("copy"):
                    factory.copy(self.settings)
                system_dirs["scan_dir"] = str(settings.dirs.data)
                components.extend(factory.scan(self._components, **system_dirs, flat=systems_flat))
            else:
                components.append(factory.load(self._components, **system_dirs, flat=systems_flat))

        except ConfigurationUnavailableError as e:
            self._logger.warning(str(e))

        if not self._components.has_type(System) and settings.dirs.data.is_default():
            components += self._components.load(configs_dir=settings.dirs.conf, configure=False, sort=False)

        if len(components) > 0:
            self._components.configure(components)
        if self._has_interface():
            self._interface.configure(settings.get_member(Interface.TYPE))

    # noinspection PyTypeChecker
    @property
    def settings(self) -> Settings:
        return self.configs

    @property
    def interface(self) -> Interface:
        return self._interface

    def _has_interface(self) -> bool:
        return self._interface is not None and self._interface.is_enabled()

    def main(self) -> None:
        action = self.settings["action"]
        try:
            if action == "run":
                with self:
                    self.run(
                        start=self.settings.get_date("start", default=None),
                        end=self.settings.get_date("end", default=None),
                    )
            elif action == "start":
                with self:
                    self.start()

            elif action == "simulate":
                with self:
                    self.simulate(
                        start=self.settings.get_date("start", default=None),
                        end=self.settings.get_date("end", default=None),
                    )

            elif action == "rotate":
                self.rotate(full=self.settings.get_bool("full"))

            elif action == "replicate":
                self.replicate(full=self.settings.get_bool("full"), force=self.settings.get_bool("force"))

        except Exception as e:
            self._logger.warning(f"Error during '{action}': {str(e)}")
            self._logger.exception(e)
            exit(1)

    def start(self, wait: bool = True) -> None:
        _has_interface = self._has_interface()
        if _has_interface:
            wait = False
        super().start(wait)

        if _has_interface:
            self._interface.start()

    # noinspection PyUnresolvedReferences, PyProtectedMember, PyShadowingBuiltins
    def simulate(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        **kwargs,
    ) -> None:
        simulation = self.settings.get_member("simulation", defaults={"data": {"include": True}})

        timezone = simulation.get("timezone", None)
        if start is None:
            start = simulation.get_date("start", default=None, timezone=timezone)
        if end is None:
            end = simulation.get_date("end", default=None, timezone=timezone)

        if start is None or end is None or end < start:
            self._logger.error("Invalid settings missing specified simulation period")
            sys.exit(1)

        slice = simulation.get_bool("slice", default=True)
        freq = simulation.get("freq", default=None)

        database_id = simulation.get("database", default="results")
        database_configs = simulation.get_member("databases", defaults={})
        if database_id == "results" and "results" not in database_configs:
            database_configs["results"] = {
                "type": "tables",
                "file": ".results.h5",
                "compression_level": 9,
                "compression_lib": "zlib",
            }
        self.connectors.load(database_configs)

        database = self.connectors.get(database_id)
        if not isinstance(database, Database):
            raise DatabaseException(database, f"Invalid results cache type '{type(database)}'")

        error = False
        summary = []
        systems = self.components.get_all(System)
        for system in systems:
            self._logger.info(f"Starting simulation of system '{system.name}': {system.id}")
            if slice and freq is not None and start + to_timedelta(freq) < end:
                slices = slice_range(start, end, timezone=system.location.timezone, freq=freq)
            else:
                slices = [(start, end)]

            with Results(system, database, simulation.get_member("data"), total=len(slices)) as results:
                results.durations.start("Simulation")
                try:
                    for slice_start, slice_end in slices:
                        slice_prior = results.data.tail(1) if not results.data.empty else None
                        results.submit(
                            system.simulate,
                            slice_start,
                            slice_end,
                            slice_prior,
                            **kwargs,
                        )
                        results.progress.update()

                    results.durations.stop("Simulation")
                    self._logger.debug(
                        f"Finished simulation of system '{system.name}' in {results.durations['Simulation']} minutes"
                    )

                    self._logger.debug(f"Starting evaluation of system '{system.name}': {system.id}")
                    results.durations.start("Evaluation")
                    results_data = system.evaluate(results)
                    results.report()

                    # TODO: Call evaluations from configs
                    if "errors" in results_data.columns.get_level_values(0):
                        self._logger.debug(f"Starting evaluation of {len(results_data['errors'].columns)} errors")

                    results.durations.stop("Evaluation")
                    self._logger.debug(
                        f"Finished evaluation of system '{system.name}' in {results.durations['Evaluation']} minutes"
                    )

                    summary.append(results.to_frame())

                except Exception as e:
                    error = True
                    self._logger.error(f"Error simulating system {system.name}: {str(e)}")
                    self._logger.debug("%s: %s", type(e).__name__, traceback.format_exc())
                    results.durations.complete()
                    results.progress.complete(
                        status="error",
                        message=str(e),
                        error=type(e).__name__,
                        trace=traceback.format_exc(),
                    )

        if not error and len(summary) > 1:
            try:
                from lories.io import excel

                excel_file = str(self.configs.dirs.data.joinpath("summary.xlsx"))
                excel.write(excel_file, "Summary", pd.concat(summary, axis="index"))

            except ImportError:
                pass



# -*- coding: utf-8 -*-
"""
lories.core.register.registry
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import builtins
from typing import Callable, Collection, Dict, Generic, Mapping, Optional, Type

from lories._core._registrator import Registrator  # noqa
from lories.core.register.registration import Registration, RegistrationError, RegistrationFactory

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import get_args

except ImportError:
    from typing_extensions import get_args


# noinspection PyShadowingBuiltins
class Registry(Generic[Registrator]):
    __types: Mapping[str, Registration[Registrator]]

    def __init__(self) -> None:
        self.__types: Dict[str, Registration[Registrator]] = {}

    def register(
        self,
        cls: Type[Registrator],
        key: str,
        *alias: str,
        factory: Optional[RegistrationFactory] = None,
        replace: bool = False,
    ) -> None:
        if not isinstance(key, str):
            raise RegistrationError(f"Invalid '{builtins.type(key)}' registration type: {key}")
        key = key.lower()
        type = self._get_generic_type()
        if not issubclass(cls, type):
            raise ValueError(f"Can only register {type} types")
        if self.has_type(key) and not replace:
            raise RegistrationError(
                f"Registration '{key}' does already exist: "
                f"{next(t for t in self.__types.values() if key == t.key).name}"
            )
        registration = Registration[Registrator](cls, key, *alias, factory=factory)
        self.__types[key] = registration

    # noinspection PyShadowingBuiltins
    def filter(
        self,
        filter: Callable[[Registration[Registrator]], bool],
    ) -> Collection[Registration[Registrator]]:
        return [c for c in self.__types.values() if filter(c)]

    def get_types(self) -> Collection[str]:
        return self.__types.keys()

    def has_type(self, type: str) -> bool:
        return any(t.is_type(type) for t in self.__types.values())

    def from_type(self, type: str) -> Registration[Registrator]:
        for registration in self.__types.values():
            if registration.is_type(type):
                return registration
        raise RegistrationError(f"Registration '{type}' does not exist")

    # noinspection PyTypeChecker, PyUnresolvedReferences
    def _get_generic_type(self) -> Type[Registrator]:
        return get_args(self.__orig_class__)[0]



# -*- coding: utf-8 -*-
"""
lories._data._manager
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from typing import Any, List, Optional, Tuple, TypeAlias, TypeVar, overload

import pandas as pd
from lories._core._activator import Activator, _Activator
from lories._core._channel import Channel, _Channel
from lories._core._channels import Channels, ChannelsArgument
from lories._core._connector import Connector
from lories._core._context import _Context
from lories._core._entity import _Entity
from lories._core.typing import Timestamp

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class _DataContext(_Context[Channel]):
    TYPE: str = "data"

    @property
    @abstractmethod
    def channels(self) -> Channels: ...

    @abstractmethod
    def filter(self, *filters: Optional[Callable[[Channel], bool]]) -> Channels: ...

    # noinspection SpellCheckingInspection
    @abstractmethod
    def groupby(self, by: str) -> List[Tuple[Any, Channels]]: ...

    @abstractmethod
    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        channels: Optional[ChannelsArgument] = None,
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        pass

    @overload
    def has_logged(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> bool: ...

    @overload
    def has_logged(
        self,
        channels: ChannelsArgument,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> bool: ...

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def has_logged(
        self,
        channels: Optional[ChannelsArgument] = None,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> bool: ...

    @overload
    def read_logged(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame: ...

    @overload
    def read_logged(
        self,
        channels: ChannelsArgument,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def read_logged(
        self,
        channels: Optional[ChannelsArgument] = None,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame: ...

    @overload
    def read(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame: ...

    @overload
    def read(
        self,
        channels: ChannelsArgument,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def read(
        self,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> pd.DataFrame: ...

    @overload
    def write(
        self,
        data: pd.DataFrame,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame: ...

    @overload
    def write(
        self,
        data: pd.DataFrame,
        channels: ChannelsArgument,
        timeout: Optional[float] = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def write(
        self,
        data: pd.DataFrame,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
    ) -> None: ...

    @abstractmethod
    def to_frame(self, **kwargs) -> pd.DataFrame: ...


class _DataManager(_DataContext, _Activator, _Entity):
    # noinspection PyShadowingBuiltins
    @abstractmethod
    def activate(self, filter: Optional[Callable[[Activator], bool]] = None) -> None: ...

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def connect(
        self,
        filter: Optional[Callable[[Connector], bool]] = None,
        channels: Optional[ChannelsArgument] = None,
        timeout: Optional[float] = None,
    ) -> None: ...

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def reconnect(
        self,
        filter: Optional[Callable[[Connector], bool]] = None,
    ) -> None: ...

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def disconnect(
        self,
        filter: Optional[Callable[[Connector], bool]] = None,
    ) -> None: ...

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def deactivate(self, *_, filter: Optional[Callable[[Activator], bool]] = None) -> None: ...


DataContext = TypeVar("DataContext", bound=_DataContext[_Channel])
DataManager = TypeVar("DataManager", bound=_DataManager)
Data: TypeAlias = DataContext



# -*- coding: utf-8 -*-
"""
lories.weather
~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections.abc import Callable
from typing import Optional, Type, TypeVar

from lories.components import Component, ComponentError, register_component_type
from lories.core import Constant, ResourceError
from lories.core.activator import ActivatorMeta
from lories.core.register import Registry
from lories.location import Location, LocationUnavailableException
from lories.typing import Configurations, ContextArgument


# noinspection PyShadowingBuiltins
def register_weather_type(
    type: str,
    *alias: str,
    factory: Callable[[ContextArgument, Optional[Configurations]], WeatherType] = None,
    replace: bool = False,
) -> Callable[[Type[WeatherType]], Type[WeatherType]]:
    # noinspection PyShadowingNames
    def _register(cls: Type[WeatherType]) -> Type[WeatherType]:
        registry.register(cls, type, *alias, factory=factory, replace=replace)
        return cls

    return _register


class WeatherMeta(ActivatorMeta):
    def __call__(cls, context: ContextArgument, configs: Configurations, **kwargs) -> WeatherType:
        _type = configs.get("type", default="default").lower()
        _cls = cls._get_class(_type)
        if cls != _cls:
            return _cls(context, configs, **kwargs)

        return super().__call__(context, configs, **kwargs)

    # noinspection PyMethodMayBeStatic, PyShadowingBuiltins
    def _get_class(cls: Type[WeatherType], type: str) -> Type[WeatherType]:
        if type in ["default", "virtual"]:
            return cls
        elif registry.has_type(type):
            registration = registry.from_type(type)
            return registration.type

        raise ResourceError(f"Unknown weather type '{type}'")


# noinspection SpellCheckingInspection
@register_component_type("weather")
class Weather(Component, metaclass=WeatherMeta):
    GHI = Constant(float, "ghi", "Global Horizontal Irradiance", "W/m2")
    DNI = Constant(float, "dni", "Direct Normal Irradiance", "W/m2")
    DHI = Constant(float, "dhi", "Diffuse Horizontal Irradiance", "W/m2")
    TEMP_AIR = Constant(float, "temp_air", "Air Temperature", "°C")
    TEMP_FELT = Constant(float, "temp_felt", "Felt Temperature", "°C")
    TEMP_DEW_POINT = Constant(float, "dew_point", "Dewpoint Temperature", "°C")
    HUMIDITY_REL = Constant(float, "humidity_relative", "Relative Humidity", "%")
    PRESSURE_SEA = Constant(float, "pressure_sea", "Atmospheric Pressure", "hPa")
    WIND_SPEED = Constant(float, "wind_speed", "Wind Speed", "km/h")
    WIND_SPEED_GUST = Constant(float, "wind_speed_gust", "Wind Gust Speed", "km/h")
    WIND_DIRECTION = Constant(float, "wind_direction", "Wind Direction", "°")
    WIND_DIRECTION_GUST = Constant(float, "wind_direction_gust", "Wind Gust Direction", "°")
    CLOUD_COVER = Constant(float, "cloud_cover", "Total Cloud Cover", "%")
    CLOUDS_LOW = Constant(float, "clouds_low", "Low Cloud Cover", "%")
    CLOUDS_MID = Constant(float, "clouds_mid", "Medium Cloud Cover", "%")
    CLOUDS_HIGH = Constant(float, "clouds_high", "High Cloud Cover", "%")
    SUNSHINE = Constant(int, "sunshine", "Sunshine Duration", "min")
    VISIBILITY = Constant(int, "visibility", "Visibility", "m")
    PRECIPITATION = Constant(float, "precipitation", "Precipitation", "mm")
    PRECIPITATION_CONV = Constant(float, "precipitation_convective", "Precipitation Convective", "mm")
    PRECIPITATION_PROB = Constant(float, "precipitation_probability", "Precipitation Probability", "%")
    PRECIPITABLE_WATER = Constant(float, "precipitable_water", "Precipitable Water", "cm")
    SNOW_FRACTION = Constant(float, "snow_fraction", "Snow Fraction", "1/0")

    location: Location

    # noinspection PyProtectedMember
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self.localize(configs.get_member(Location.TYPE, defaults={}))

    # noinspection PyUnresolvedReferences
    def localize(self, configs: Configurations) -> None:
        if configs.enabled and all(k in configs for k in ["latitude", "longitude"]):
            self.location = Location(
                configs.get_float("latitude"),
                configs.get_float("longitude"),
                timezone=configs.get("timezone", default="UTC"),
                altitude=configs.get_float("altitude", default=None),
                country=configs.get("country", default=None),
                state=configs.get("state", default=None),
            )
        else:
            try:
                self.location = self.context.location
                if not isinstance(self.location, Location):
                    raise ComponentError(self, f"Invalid location type '{type(self.location)}'")
            except (LocationUnavailableException, AttributeError):
                raise ComponentError(self, "Missing location")


WeatherType = TypeVar("WeatherType", bound=Weather)

registry = Registry[Weather]()



# -*- coding: utf-8 -*-
"""
lories.components.weather
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from . import weather  # noqa: F401
from .weather import (  # noqa: F401
    Weather,
    register_weather_type,
    registry,
)

from . import forecast  # noqa: F401
from .forecast import WeatherForecast  # noqa: F401

from . import provider  # noqa: F401
from .provider import WeatherProvider  # noqa: F401

from . import dwd  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.application.interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections.abc import Callable
from typing import Optional, Type, TypeVar

from lories._core._configurations import Configurations  # noqa
from lories._core._data import DataContext, _DataContext, _DataManager  # noqa
from lories.core import ConfigurationError, Registry, ResourceError
from lories.core.configs.configurator import Configurator, ConfiguratorMeta


# noinspection PyShadowingBuiltins
def register_interface_type(
    type: str,
    *alias: str,
    factory: Callable[[DataContext, Configurations], InterfaceType] = None,
    replace: bool = False,
) -> Callable[[Type[InterfaceType]], Type[InterfaceType]]:
    # noinspection PyShadowingNames
    def _register(cls: Type[InterfaceType]) -> Type[InterfaceType]:
        registry.register(cls, type, *alias, factory=factory, replace=replace)
        return cls

    return _register


class InterfaceMeta(ConfiguratorMeta):
    def __call__(cls, context: DataContext, configs: Configurations, **kwargs) -> InterfaceType:
        global _instance

        _type = configs.get("type", default="default").lower()
        _cls = cls._get_class(_type)
        if cls != _cls:
            return _cls(context, configs, **kwargs)
        if _instance is None:
            _instance = super().__call__(context, configs, **kwargs)
        return _instance

    # noinspection PyMethodMayBeStatic, PyShadowingBuiltins
    def _get_class(cls: Type[InterfaceType], type: str) -> Type[InterfaceType]:
        if type == "default":
            return cls
        elif registry.has_type(type):
            registration = registry.from_type(type)
            return registration.type

        raise InterfaceException(f"Unknown interface type '{type}'")


class Interface(Configurator, metaclass=InterfaceMeta):
    TYPE: str = "interface"

    __context: _DataContext

    # noinspection PyUnresolvedReferences
    def __init__(self, context: DataContext, configs: Configurations, **kwargs) -> None:
        super().__init__(configs, **kwargs)
        self.__context = self._assert_context(context)

    @classmethod
    def _assert_context(cls, context: DataContext) -> DataContext:
        if context is None or not isinstance(context, _DataManager):
            raise ResourceError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return context

    @classmethod
    def _assert_configs(cls, configs: Optional[Configurations]) -> Optional[Configurations]:
        if configs is None:
            raise ConfigurationError(f"Invalid '{cls.__name__}' configurations: {type(configs)}")
        return super()._assert_configs(configs)

    @property
    def context(self) -> DataContext:
        return self.__context

    def start(self, *args, **kwargs) -> None:
        pass


class InterfaceException(ResourceError):
    """
    Raise if an error occurred accessing the interface.

    """


InterfaceType = TypeVar("InterfaceType", bound=Interface)

# FIXME: Add global variable typing again once Python >= 3.9 is a requirement
# _instance: Optional[InterfaceType] = None
_instance = None
registry = Registry[Interface]()



# -*- coding: utf-8 -*-
"""
lories.data.converters.access
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories._core._component import Component  # noqa
from lories._core._converter import Converter, _Converter, _ConverterContext  # noqa
from lories._core._data import _DataManager  # noqa
from lories.core import RegistratorAccess, ResourceError
from lories.util import get_context


class ConverterAccess(_ConverterContext, RegistratorAccess[Converter]):
    # noinspection PyUnresolvedReferences
    def __init__(self, registrar: Component, **kwargs) -> None:
        context = get_context(registrar, _DataManager).converters
        super().__init__(context, registrar, **kwargs)

    # noinspection PyProtectedMember, PyShadowingBuiltins
    def _set(self, id: str, converter: Converter) -> None:
        if not isinstance(converter, _Converter):
            raise ResourceError(f"Invalid converter type: {type(converter)}")

        super()._set(id, converter)



"""
lories.typing
~~~~~~~~~~~~~


"""

from typing import TypeVar

from ..core.typing import (
    Timestamp,
    Timezone,
    Entity,
    Context,
    ContextArgument,
    Resource,
    Resources,
    ResourcesArgument,
    Channel,
    Channels,
    ChannelsArgument,
    Converter,
    Converters,
    Listener,
    Configurations,
    Configurator,
    Registrator,
    Activator,
    Connector,
    Connectors,
    Database,
    Data,
    Component,
    Components,
)

from ..location import Location as _Location

from ..components import Weather as _Weather
from ..components import Tariff as _Tariff

from ..system import System as _System


Location = TypeVar("Location", bound=_Location)
Weather = TypeVar("Weather", bound=_Weather)
Tariff = TypeVar("Tariff", bound=_Tariff)
System = TypeVar("System", bound=_System)

__all__ = [
    "Timestamp",
    "Timezone",
    "Entity",
    "Context",
    "ContextArgument",
    "Resource",
    "Resources",
    "ResourcesArgument",
    "Channel",
    "Channels",
    "ChannelsArgument",
    "Converter",
    "Converters",
    "Listener",
    "Configurations",
    "Configurator",
    "Registrator",
    "Activator",
    "Connector",
    "Connectors",
    "Database",
    "Data",
    "Component",
    "Components",
    "Location",
    "Weather",
    "Tariff",
    "System",
]



# -*- coding: utf-8 -*-
"""
lories.data.channels.channel
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Callable
from typing import Any, Collection, Dict, List, Optional, Type

import pandas as pd
import pytz as tz
from lories._core._channel import ChannelState, _Channel  # noqa
from lories._core._data import DataContext, DataManager, _DataContext, _DataManager  # noqa
from lories._core.typing import Timestamp  # noqa
from lories.core import Resource, ResourceError
from lories.data.channels import ChannelConnector, ChannelConverter, Channels
from lories.util import parse_freq, to_timedelta

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class Channel(_Channel, Resource):
    __context: _DataContext

    _timestamp: pd.Timestamp = pd.NaT
    _value: Optional[Any] = None
    _state: str | ChannelState = ChannelState.DISABLED

    logger: ChannelConnector
    connector: ChannelConnector
    converter: ChannelConverter

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        id: str = None,
        key: str = None,
        name: str = None,
        type: str | Type = None,
        context: DataManager = None,
        converter: ChannelConverter = None,
        connector: Optional[ChannelConnector] = None,
        logger: Optional[ChannelConnector] = None,
        **configs: Any,
    ) -> None:
        super().__init__(id=id, key=key, name=name, type=type, **configs)
        self.__context = self._assert_context(context)
        self.converter = self._assert_converter(converter)
        self.connector = self._assert_connector(connector)
        self.logger = self._assert_connector(logger)

    @classmethod
    def _assert_context(cls, context: DataManager) -> DataManager:
        if context is None or not isinstance(context, _DataManager):
            raise ResourceError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return context

    @classmethod
    def _assert_converter(cls, converter: Optional[ChannelConverter]) -> ChannelConverter:
        if converter is None or not isinstance(converter, ChannelConverter):
            raise ResourceError(f"Invalid channel converter: {type(converter)}")
        return converter

    @classmethod
    def _assert_connector(cls, connector: Optional[ChannelConnector]) -> ChannelConnector:
        if connector is None:
            connector = ChannelConnector()
        elif not isinstance(connector, ChannelConnector):
            raise ResourceError(f"Invalid channel connector: {type(connector)}")
        return connector

    def _get_attrs(self) -> List[str]:
        return [
            *super()._get_attrs(),
            "converter",
            "connector",
            "logger",
            "value",
            "state",
            "timestamp",
        ]

    # noinspection PyShadowingBuiltins
    def _get_vars(self) -> Dict[str, Any]:
        vars = super()._get_vars()
        vars["value"] = self.value
        vars["state"] = self.state
        vars["timestamp"] = self.timestamp
        return vars

    # noinspection PyShadowingBuiltins
    def __repr__(self) -> str:
        vars = OrderedDict(key=self.id)
        if self.is_valid():
            vars["value"] = str(self.value)
        else:
            vars["state"] = str(self.state)
        vars["timestamp"] = str(self.timestamp)
        return f"{type(self).__name__}({', '.join(f'{k}={v}' for k, v in vars.items())})"

    @property
    def freq(self) -> Optional[str]:
        freq = self.get(next((k for k in ["freq", "frequency", "resolution"] if k in self), None), default=None)
        if freq is not None:
            freq = parse_freq(freq)
        return freq

    @property
    def timedelta(self) -> Optional[pd.Timedelta]:
        return to_timedelta(self.freq)

    @property
    def timestamp(self) -> pd.Timestamp:
        return self._timestamp

    @property
    def value(self) -> Optional[Any]:
        return self._value

    @value.setter
    def value(self, value) -> None:
        self._set(pd.Timestamp.now(tz.UTC).floor(freq="s"), value, ChannelState.VALID)

    @property
    def state(self) -> ChannelState | str:
        return self._state

    @state.setter
    def state(self, state) -> None:
        self._set(pd.Timestamp.now(tz.UTC).floor(freq="s"), None, state)

    def is_valid(self) -> bool:
        return self._is_valid(self.value, self.state)

    @staticmethod
    def _is_valid(value: Any, state: ChannelState) -> bool:
        return state == ChannelState.VALID and not Channel._is_empty(value)

    @staticmethod
    def _is_empty(value: Any) -> bool:
        if isinstance(value, Collection) and not isinstance(value, str) and not isinstance(value, bytes):
            return any(pd.isna(value))
        return pd.isna(value)

    def set(
        self,
        timestamp: pd.Timestamp,
        value: Any,
        state: Optional[str | ChannelState] = ChannelState.VALID,
    ) -> None:
        value = self.converter(value)
        self._set(timestamp, value, state)

    # noinspection PyUnresolvedReferences
    def _set(
        self,
        timestamp: pd.Timestamp,
        value: Optional[Any],
        state: str | ChannelState,
    ) -> None:
        if not isinstance(timestamp, pd.Timestamp):
            raise ResourceError(f"Expected pandas Timestamp for '{self.id}', not: {type(value)}")
        self._timestamp = timestamp

        if self._is_empty(value) and state == ChannelState.VALID:
            raise ResourceError(f"Invalid value for valid state '{self.id}': {value}")
        self._value = value
        self._state = state

    # noinspection PyShadowingBuiltins, PyProtectedMember
    def _update(
        self,
        converter: Optional[Dict[str, Any] | str] = None,
        connector: Optional[Dict[str, Any] | str] = None,
        logger: Optional[Dict[str, Any] | str] = None,
        **configs: Any,
    ) -> None:
        if converter is not None:
            converter = Channel._build_member(converter, "converter")
            self.converter._update(**converter)
        if connector is not None:
            connector = Channel._build_member(connector, "connector")
            self.connector._update(**connector)
        if logger is not None:
            logger = Channel._build_member(logger, "connector")
            self.logger._update(**logger)
        super()._update(**configs)

    def _copy_args(self) -> Dict[str, Any]:
        arguments = super()._copy_args()
        arguments["context"] = self.__context
        # arguments["processors"] = self.processors.copy()
        arguments["converter"] = self.converter.copy()
        arguments["connector"] = self.connector.copy()
        arguments["logger"] = self.logger.copy()

        return arguments

    def copy(self) -> Channel:
        channel = super().copy()
        channel._timestamp = self._timestamp
        channel._value = self._value
        channel._state = self._state
        return channel

    def to_list(self) -> Channels:
        return Channels([self])

    def to_configs(self) -> Dict[str, Any]:
        configs = super().to_configs()
        configs["converter"] = self.converter.to_configs()
        configs["connector"] = self.connector.to_configs()
        configs["logger"] = self.logger.to_configs()
        return configs

    def to_series(self, state: bool = False) -> pd.Series:
        if not self.is_valid() and state:
            return pd.Series(data=[self.state], index=[self.timestamp], name=self.key)

        return self.converter.to_series(self.value, self.timestamp, name=self.key)

    # noinspection PyProtectedMember
    def from_logger(self) -> Channel:
        channel = self.copy()
        channel._update(**self.logger._copy_configs())
        return channel

    # noinspection PyShadowingBuiltins
    def has_logger(self, *ids: Optional[str]) -> bool:
        return self.logger.enabled and (any(self.logger.id == id for id in ids) if len(ids) > 0 else True)

    # noinspection PyShadowingBuiltins
    def has_connector(self, id: Optional[str] = None) -> bool:
        return self.connector.enabled and (self.connector.id == id if id is not None else True)

    # noinspection PyUnresolvedReferences
    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        self.__context.register(function, self, how=how, unique=unique)

    # noinspection PyUnresolvedReferences
    def read(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        return self.__context.read(self.to_list(), start, end)

    # noinspection PyUnresolvedReferences
    def write(self, data: pd.DataFrame | pd.Series | Any) -> None:
        if data is None:
            raise ResourceError(f"Invalid data to write '{self.id}': {data}")
        if isinstance(data, pd.Series):
            data.name = self.id
            data = data.to_frame()
        elif not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(index=[pd.Timestamp.now(tz.UTC).floor(freq="s")], data=[data], columns=[self.id])

        self.__context.write(data, self.to_list())



# -*- coding: utf-8 -*-
"""
lories.data.converters.converter
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
import json
from typing import Any, Generic, Optional, Type, TypeVar

import pandas as pd
import pytz as tz
from lories._core import _Channel, _Channels, _Converter  # noqa
from lories.core import Registrator
from lories.data.converters.errors import ConversionError
from lories.data.validation import validate_index
from lories.util import is_bool, is_float, is_int, to_bool, to_date, to_float, to_int

T = TypeVar("T", bound=Any)


# noinspection PyAbstractClass
class Converter(_Converter, Registrator, Generic[T]):
    def to_str(self, value: T | pd.Series) -> str:
        return self.to_json(value)

    # noinspection PyMethodMayBeStatic
    def to_json(self, data: T | pd.Series | pd.DataFrame) -> str:
        if issubclass(type(data), (pd.Series, pd.DataFrame)):
            if issubclass(type(data), pd.Series):
                data = data.to_frame()
            columns = data.columns
            data[_Channel.TIMESTAMP] = data.index
            return data[[_Channel.TIMESTAMP, *columns]].to_json(orient="records")
        return json.dumps(data)

    # noinspection PyMethodMayBeStatic
    def to_series(self, value: T, timestamp: Optional[pd.Timestamp] = None, name: Optional[str] = None) -> pd.Series:
        if timestamp is None:
            timestamp = pd.Timestamp.now(tz=tz.UTC)
        if isinstance(value, pd.Series):
            series = value
            series.name = name
        else:
            series = pd.Series(index=[timestamp], data=[value], name=name)

        series = validate_index(series)
        series.index.name = _Channel.TIMESTAMP
        if series.index.tzinfo is None:
            self._logger.warning(
                f"UTC will be presumed for timestamps, as DatetimeIndex are expected to be tz-aware: {series}"
            )
            series.index = series.index.tz_localize(tz.UTC)
        return series

    # noinspection PyProtectedMember
    def from_frame(self, data: pd.DataFrame, channels: _Channels) -> pd.DataFrame:
        converted_data = []
        for channel in channels:
            channel_data = data[channel.id].dropna() if channel.id in data.columns else None
            if channel_data is None or channel_data.empty:
                converted_data.append(pd.Series(name=channel.id))
                continue
            elif not channel_data.apply(self.is_dtype).all():
                converted_data.append(pd.Series(name=channel.id))
                self._logger.warning(f"Unable to convert values for channel '{channel.id}': {channel_data.values}")
                continue
            try:
                converted_data.append(self.from_series(channel_data, channel))
            except TypeError:
                raise ConversionError(f"Expected str or {self.dtype}, not: {type(data)}")
        if len(converted_data) == 0:
            return pd.DataFrame(columns=[c.id for c in channels])
        return pd.concat(converted_data, axis="columns")

    # noinspection PyProtectedMember, PyUnresolvedReferences
    def from_series(self, data: pd.Series, channel: _Channel) -> pd.Series:
        try:
            converter_args = channel.converter._get_configs()
            converted_data = data.apply(self.convert, **converter_args)
            return converted_data.apply(self.to_dtype, **converter_args)
        except TypeError:
            raise ConversionError(f"Expected str or {self.dtype}, not: {type(data)}")

    # noinspection PyMethodMayBeStatic, PyUnusedLocal
    def convert(self, value: Any, **kwargs) -> Optional[T]:
        return value


# noinspection PyAbstractClass, PyMethodMayBeStatic
class _NumberConverter(Converter[T]):
    def scale(self, value: T, factor: Optional[T], invert: bool = False, **kwargs) -> T:
        if value is not None and factor is not None:
            if invert:
                value /= factor
            else:
                value *= factor
        return self.to_dtype(value, **kwargs)

    # noinspection PyProtectedMember, PyUnresolvedReferences
    def from_series(self, data: pd.Series, channel: _Channel) -> pd.Series:
        try:
            factor = to_float(channel.get("scale", default=None))
            converter_args = channel.converter._get_configs()
            converted_data = data.apply(self.convert, **converter_args)
            return converted_data.apply(self.scale, args=(factor,), **converter_args)
        except TypeError:
            raise ConversionError(f"Expected str or {self.dtype}, not: {type(data)}")


# noinspection PyMethodMayBeStatic
class DatetimeConverter(Converter[dt.datetime]):
    dtype: Type[dt.datetime] = dt.datetime

    def is_dtype(self, value: str | dt.datetime) -> bool:
        return isinstance(value, (str, self.dtype))

    def to_dtype(self, value: str | dt.datetime, **_) -> Optional[dt.datetime]:
        return to_date(value)


# noinspection PyMethodMayBeStatic
class TimestampConverter(DatetimeConverter):
    dtype: Type[pd.Timestamp] = pd.Timestamp

    def is_dtype(self, value: str | pd.Timestamp) -> bool:
        return isinstance(value, (str, self.dtype))

    def to_dtype(self, value: str | pd.Timestamp, **_) -> Optional[pd.Timestamp]:
        return to_date(value)


# noinspection PyMethodMayBeStatic
class StringConverter(Converter[str]):
    dtype: Type[str] = str

    def is_dtype(self, value: str) -> bool:
        return isinstance(value, str)

    def to_dtype(self, value: Any, **_) -> Optional[str]:
        return str(value)


# noinspection PyMethodMayBeStatic
class FloatConverter(_NumberConverter[float]):
    dtype: Type[float] = float

    def is_dtype(self, value: str | float) -> bool:
        return is_float(value)

    def to_dtype(self, value: str | float, decimals: Optional[int] = None, **_) -> Optional[float]:
        value = to_float(value)
        if decimals is not None:
            value = round(value, decimals)
        return value


# noinspection PyMethodMayBeStatic
class IntConverter(_NumberConverter[int]):
    dtype: Type[int] = int

    def is_dtype(self, value: str | int) -> bool:
        return is_int(value)

    def to_dtype(self, value: str | int, **_) -> Optional[int]:
        return to_int(value)


# noinspection PyMethodMayBeStatic
class BoolConverter(Converter[bool]):
    dtype: Type[bool] = bool

    def is_dtype(self, value: str | bool) -> bool:
        return is_bool(value) or is_int(value)

    def to_dtype(self, value: str | bool, **_) -> Optional[bool]:
        return to_bool(value)


# noinspection PyMethodMayBeStatic
class BytesConverter(Converter[bytes]):
    dtype: Type[bytes] = bytes

    def is_dtype(self, value: str | bytes) -> bool:
        return isinstance(value, bytes)

    def to_dtype(self, value: str | bytes, **_) -> Optional[bytes]:
        if isinstance(value, bytes):
            return value
        elif isinstance(value, str):
            return value.encode()
        return None



# -*- coding: utf-8 -*-
"""
lories.core.register.registrator
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import inspect
import sys
from collections import OrderedDict
from collections.abc import Callable
from typing import Any, Dict, Optional, TypeVar

from lories._core import _Context, _Entity, _Resource, _Resources  # noqa
from lories._core._registrator import RegistratorContext, _Registrator, _RegistratorContext  # noqa
from lories.core.configs import Configurations, Configurator
from lories.core.errors import ResourceError


class Registrator(_Registrator, Configurator):
    __context: RegistratorContext

    # noinspection PyShadowingBuiltins, PyProtectedMember, PyUnresolvedReferences, PyUnusedLocal
    def __init__(
        self,
        context: RegistratorContext,
        configs: Optional[Configurations] = None,
        id: Optional[str] = None,
        key: Optional[str] = None,
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        context = self._assert_context(context)

        name = self._build_name(name, configs=configs)
        key = self._build_key(key, configs=configs)
        id = self._build_id(id, key, configs=configs, context=context)

        if configs is None:
            configs = Configurations(f"{key}.conf", context.configs.dirs)
        super().__init__(
            configs=configs,
            id=id,
            key=key,
            name=name,
            **kwargs,
        )
        self.__context = context

    @classmethod
    def _assert_context(
        cls,
        context: RegistratorContext,
    ) -> RegistratorContext:
        if context is None or not isinstance(context, (_RegistratorContext, _Registrator)):
            raise ResourceError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return context

    # noinspection PyShadowingBuiltins
    def _convert_vars(self, convert: Callable[[Any], str] = str) -> Dict[str, str]:
        vars = self._get_vars()
        values = OrderedDict()
        try:
            id = vars.pop("id", self.id)
            key = vars.pop("key", self.key)
            if id != key:
                values["key"] = id
            values["key"] = key
            values["name"] = vars.pop("name", self.name)
        except (ResourceError, AttributeError):
            # Abstract properties are not yet instanced
            pass

        def is_from_framework(value: Any) -> bool:
            if isinstance(value, (Configurator, _Context, _Entity, _Resource, _Resources)):
                return True
            module = inspect.getmodule(value)
            if module is None:
                return False
            base, *_ = module.__name__.partition(".")
            return sys.modules[base] == "lories"

        values.update({k: str(v) if not is_from_framework(v) else convert(v) for k, v in vars.items()})
        values["enabled"] = str(self.is_enabled())
        values["configured"] = str(self.is_configured())
        values["configs"] = convert(self.configs)
        values["context"] = convert(self.context)
        return values

    @property
    def context(self) -> RegistratorContext:
        return self.__context

    # noinspection PyProtectedMember, PyTypeChecker, PyShadowingBuiltins
    def duplicate(self, context: Optional[RegistratorContext] = None, **changes):
        if context is None:
            context = self.__context
        return super().duplicate(context=context, **changes)


RegistratorType = TypeVar("RegistratorType", bound=Registrator)



# -*- coding: utf-8 -*-
"""
lories.core.configs.errors
~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories.core.errors import ResourceError, ResourceUnavailableError


class ConfigurationError(ResourceError):
    """
    Raise if a configuration is invalid.

    """


class ConfigurationUnavailableError(ResourceUnavailableError, ConfigurationError):
    """
    Raise if a configuration file can not be found.

    """



# -*- coding: utf-8 -*-
"""
lories.connectors
~~~~~~~~~~~~~~~~~


"""

from .._core import (  # noqa: F401
    ConnectType,
)

from .errors import (  # noqa: F401
    ConnectorError,
    ConnectorUnavailableError,
    ConnectionError,
    DatabaseException,
    DatabaseUnavailableException,
)

from . import access  # noqa: F401
from .access import ConnectorAccess  # noqa: F401

from . import context  # noqa: F401
from .context import (  # noqa: F401
    ConnectorContext,
    register_connector_type,
    registry,
)

from . import connector  # noqa: F401
from .connector import Connector  # noqa: F401

from ..data import database  # noqa: F401
from ..data.database import Database  # noqa: F401

from ..data import databases  # noqa: F401
from ..data.databases import Databases  # noqa: F401

import importlib

CONNECTORS = [
    "virtual",
    "csv",
    "sql",
    "influx",
    "tables",
    "cameras",
    "serial",
    "modbus",
    "revpi",
    "entsoe",
]

for import_connector in CONNECTORS:
    try:
        importlib.import_module(f".{import_connector}", "lories.connectors")

    except ModuleNotFoundError:
        # TODO: Implement meaningful logging here
        pass

del importlib



# -*- coding: utf-8 -*-
"""
lories.components.tariff.provider
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories.components.tariff import Tariff
from lories.typing import Configurations


# noinspection SpellCheckingInspection
class TariffProvider(Tariff):
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self.data.add(Tariff.PRICE_IMPORT, aggregate="mean", connector=None)
        self.data.add(Tariff.PRICE_EXPORT, aggregate="mean", connector=None)



# -*- coding: utf-8 -*-
"""
lories._core
~~~~~~~~~~~~


"""

from ._constant import _Constant  # noqa: F401

from ._entity import _Entity  # noqa: F401
from ._context import _Context  # noqa: F401

from ._resource import _Resource  # noqa: F401
from ._resources import _Resources  # noqa: F401

from ._channel import (  # noqa: F401
    _Channel,
    ChannelState,
)
from ._channels import _Channels  # noqa: F401

from ._registrator import (  # noqa: F401
    _Registrator,
    _RegistratorContext,
)

from ._converter import (  # noqa: F401
    _Converter,
    _ConverterContext,
)

from ._listener import _Listener  # noqa: F401

from ._configurations import _Configurations  # noqa: F401
from ._configurator import _Configurator  # noqa: F401

from ._activator import _Activator  # noqa: F401

from ._connector import (  # noqa: F401
    _Connector,
    _ConnectorContext,
    ConnectType,
)
from ._database import _Database  # noqa: F401

from ._data import (  # noqa: F401
    _DataContext,
    _DataManager,
)

from ._component import (  # noqa: F401
    _Component,
    _ComponentContext,
)



# -*- coding: utf-8 -*-
"""
lories.connectors.tables
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import os
import re
from typing import Optional, Sequence

import pandas as pd
from lories.connectors import ConnectionError, Database, register_connector_type
from lories.typing import Configurations, Resources, Timestamp
from pandas import HDFStore


@register_connector_type("tables", "hdfstore")
class HDFDatabase(Database):
    __store: HDFStore = None

    _store_dir: str
    _store_path: str

    _mode: str = "a"

    _columns_unique: bool = True

    _compression_level: int | None = None
    _compression_lib = None

    # noinspection PyTypeChecker
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        store_dir = configs.get("path", default=None)
        if store_dir is not None:
            if "~" in store_dir:
                store_dir = os.path.expanduser(store_dir)
            if not os.path.isabs(store_dir):
                store_dir = os.path.join(configs.dirs.data, store_dir)
        else:
            store_dir = configs.dirs.data

        store_path = configs.get("path", default=None)
        if store_path is not None:
            if not os.path.isabs(store_path):
                store_path = os.path.join(store_dir, store_path)
            else:
                store_dir = os.path.dirname(store_path)
        else:
            store_file = configs.get("file", default=".store.h5")
            store_path = os.path.join(store_dir, store_file)
        self._store_dir = store_dir
        self._store_path = store_path

        self._mode = configs.get("mode", default="a")
        self._columns_unique = configs.get_bool("columns_unique", default=False)
        self._compression_level = configs.get_int("compression_level", None)
        self._compression_lib = configs.get("compression_lib", None)

    def is_connected(self) -> bool:
        return self.__store is not None and self.__store.is_open

    def connect(self, resources: Resources) -> None:
        super().connect(resources)
        if not os.path.isdir(self._store_dir):
            os.makedirs(self._store_dir, exist_ok=True)
        try:
            self.__store = HDFStore(
                self._store_path,
                mode=self._mode,
                complevel=self._compression_level,
                complib=self._compression_lib,
            )
            self.__store.open()

        except IOError as e:
            raise ConnectionError(self, str(e))

    def disconnect(self) -> None:
        super().disconnect()
        if self.__store is not None:
            self.__store.close()

    # noinspection PyTypeChecker, PyUnresolvedReferences
    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        data = []
        try:
            for group, group_resources in resources.groupby("group"):
                group_key = _format_key(group)
                if group_key not in self.__store:
                    continue

                group_data = self.__store.select(
                    group_key,
                    where=_build_where(start, end),
                    columns=self.__build_columns(group_resources),
                )
                group_data = self.__extract_data(group_resources, group_data)
                if not group_data.empty:
                    data.append(group_data)
        except IOError as e:
            raise ConnectionError(self, str(e))

        if len(data) == 0:
            return pd.DataFrame()
        data = sorted(data, key=lambda d: min(d.index))
        return pd.concat(data, axis="columns")

    # noinspection PyTypeChecker, PyUnresolvedReferences
    def read_first(self, resources: Resources) -> Optional[pd.DataFrame]:
        data = []
        try:
            for group, group_resources in resources.groupby("group"):
                group_key = _format_key(group)
                if group_key not in self.__store:
                    continue

                group_data = self.__store.select(group_key, stop=1, columns=self.__build_columns(group_resources))
                group_data = self.__extract_data(group_resources, group_data)
                if not group_data.empty:
                    data.append(group_data)
        except IOError as e:
            raise ConnectionError(self, str(e))

        if len(data) == 0:
            return pd.DataFrame()
        data = sorted(data, key=lambda d: min(d.index))
        return pd.concat(data, axis="columns")

    # noinspection PyTypeChecker, PyUnresolvedReferences
    def read_last(self, resources: Resources) -> Optional[pd.DataFrame]:
        data = []
        try:
            for group, group_resources in resources.groupby("group"):
                group_key = _format_key(group)
                if group_key not in self.__store:
                    continue

                group_data = self.__store.select(group_key, start=-1, columns=self.__build_columns(group_resources))
                group_data = self.__extract_data(group_resources, group_data)
                if not group_data.empty:
                    data.append(group_data)
        except IOError as e:
            raise ConnectionError(self, str(e))

        if len(data) == 0:
            return pd.DataFrame()
        data = sorted(data, key=lambda d: min(d.index))
        return pd.concat(data, axis="columns")

    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> None:
        try:
            for group, group_resources in resources.groupby("group"):
                group_key = _format_key(group)
                if group_key not in self.__store:
                    continue

                self.__store.remove(group_key)

        except IOError as e:
            raise ConnectionError(self, str(e))

    def write(self, data: pd.DataFrame) -> None:
        try:
            for group, group_resources in self.resources.filter(lambda c: c.id in data.columns).groupby("group"):
                group_key = _format_key(group)
                group_data = data[group_resources.ids].dropna(axis="index", how="all").dropna(axis="columns", how="all")

                if not self._columns_unique:
                    group_data.rename(
                        columns={r.id: r.get("column", default=r.key) for r in group_resources},
                        inplace=True,
                    )
                if group_key not in self.__store:
                    self.__store.put(group_key, group_data, format="table", encoding="UTF-8")
                else:
                    self.__store.append(group_key, group_data, format="table", encoding="UTF-8")

        except IOError as e:
            raise ConnectionError(self, str(e))

    def __build_columns(self, resources: Resources) -> Sequence[str]:
        if self._columns_unique:
            return [r.id for r in resources]
        return [r.get("column", default=r.key) for r in resources]

    def __extract_data(self, resources: Resources, data: pd.DataFrame) -> pd.DataFrame:
        data.dropna(axis="columns", how="all", inplace=True)
        if not self._columns_unique:
            return data.rename(columns={r.get("column", default=r.key): r.id for r in resources})
        return data


def _format_key(key: str) -> str:
    key = re.sub(r"\W", "/", key).replace("_", "/").lower()
    return f"/{key}"


def _build_where(
    start: Optional[Timestamp] = None,
    end: Optional[Timestamp] = None,
) -> Optional[str]:
    where = []
    if start is not None:
        where.append(f'index>=Timestamp("{start.isoformat()}")')
    if end is not None:
        where.append(f'index<=Timestamp("{end.isoformat()}")')
    return " & ".join(where) if len(where) > 0 else None



# -*- coding: utf-8 -*-
"""
lories.components.weather.dwd.provider
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Collection, Dict

import pandas as pd
from lories import Constant
from lories.components.weather import Weather, WeatherForecast, WeatherProvider, register_weather_type
from lories.components.weather.dwd import Brightsky
from lories.typing import Configurations

CHANNELS = [
    Weather.GHI,
    Weather.TEMP_AIR,
    Weather.TEMP_DEW_POINT,
    Weather.PRESSURE_SEA,
    Weather.WIND_SPEED,
    Weather.WIND_SPEED_GUST,
    Weather.WIND_DIRECTION,
    Weather.CLOUD_COVER,
    Weather.SUNSHINE,
    Weather.VISIBILITY,
    Weather.PRECIPITATION,
    Weather.PRECIPITATION_PROB,
    Constant(str, "condition", "Condition"),
    Constant(str, "icon", "Icon"),
]

CHANNEL_ADDRESS_ALIAS = {
    Weather.GHI: "solar",
    Weather.TEMP_AIR: "temperature",
    Weather.PRESSURE_SEA: "pressure_msl",
    Weather.WIND_SPEED_GUST: "wind_gust_speed",
    Weather.WIND_DIRECTION_GUST: "wind_gust_direction",
}

CHANNEL_AGGREGATE_ALIAS = {
    Weather.SUNSHINE: "sum",
    Weather.PRECIPITATION: "sum",
    Weather.WIND_SPEED_GUST: "max",
    "condition": None,
    "icon": None,
}


# noinspection SpellCheckingInspection
@register_weather_type("dwd", "brightsky")
class DeutscherWetterDienst(WeatherProvider):
    # noinspection PyUnresolvedReferences
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        connector = Brightsky(self, self.location, key="brightsky")
        self.connectors.add(connector)

        if self.forecast.is_enabled() and isinstance(self.forecast, WeatherForecast):
            self.forecast.data.add(
                key="timestamp_creation",
                name="Creation Timestamp",
                connector=connector.id,
                address="source_first_record",
                source="forecast",
                type=pd.Timestamp,
                aggregate="last",
                logger={
                    "primary": True,
                    "nullable": False,
                },
            )
            for channel in build_channels(connector=connector.id, source="forecast"):
                self.forecast.data.add(**channel)

        for channel in build_channels(connector=connector.id, source="current, historical"):
            if channel["key"] not in [Weather.PRECIPITATION_PROB]:
                self.data.add(**channel)


def build_channels(**custom: Any) -> Collection[Dict[str, Any]]:
    channels = []
    for channel in CHANNELS:
        configs = channel.to_dict()
        if channel.key in CHANNEL_ADDRESS_ALIAS:
            configs["address"] = CHANNEL_ADDRESS_ALIAS[channel.key]
        else:
            configs["address"] = channel.key
        if configs["type"] == str:  # noqa: E721
            configs["length"] = 32
        if channel.key in CHANNEL_AGGREGATE_ALIAS:
            configs["aggregate"] = CHANNEL_AGGREGATE_ALIAS[channel.key]
        else:
            configs["aggregate"] = "mean"
        configs.update(custom)
        channels.append(configs)
    return channels



# -*- coding: utf-8 -*-
"""
lories._core._entity
~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import ABC
from typing import Any, Optional, TypeVar

from lories.util import parse_name, validate_key


class _Entity(ABC):
    _id: str
    _key: str
    _name: str

    # noinspection PyProtectedMember, PyShadowingBuiltins
    def __init__(
        self,
        id: Optional[str] = None,
        key: Optional[str] = None,
        name: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        key = self._assert_key(key)
        id = self._assert_id(id, key)

        self._id = id
        self._key = key
        self._name = self._assert_name(name, key)

    def __eq__(self, other: Any) -> bool:
        return self is other

    def __hash__(self) -> int:
        return hash(id(self))

    def __copy__(self) -> Entity:
        return self.copy()

    def __replace__(self, **changes) -> Entity:
        return self.duplicate(**changes)

    # noinspection PyShadowingBuiltins
    @classmethod
    def _assert_id(cls, __id: Optional[str], __key: Optional[str]) -> str:
        if __id is None and __key is None:
            raise TypeError(f"Invalid {cls.__name__}, missing specified 'id'")
        if __id is None:
            __id = __key
        _id = ".".join(validate_key(i) for i in __id.split("."))
        if _id != __id:
            raise ValueError(f"Invalid characters in '{cls.__name__}' id: " + __id)
        return _id

    @classmethod
    def _assert_key(cls, __key: str) -> str:
        if __key is None:
            raise TypeError(f"Invalid {cls.__name__}, missing specified 'key'")
        _key = validate_key(__key)
        if _key != __key:
            raise ValueError(f"Invalid characters in '{cls.__name__}' key: " + __key)
        return _key

    @classmethod
    def _assert_name(cls, __name: Optional[str], __key: Optional[str]) -> str:
        if __name is None and __key is None:
            raise TypeError(f"Invalid {cls.__name__}, missing specified 'name'")
        if __name is None:
            __name = parse_name(__key)
        return __name

    @property
    def id(self) -> str:
        return self._id

    @property
    def key(self) -> str:
        return self._key

    @property
    def name(self) -> str:
        return self._name

    # noinspection PyUnresolvedReferences
    def copy(self) -> Entity:
        try:
            copier = super().copy
        except AttributeError:
            copier = self.duplicate
        return copier()

    # noinspection PyUnresolvedReferences, PyShadowingBuiltins
    def duplicate(
        self,
        id: Optional[str] = None,
        key: Optional[str] = None,
        name: Optional[str] = None,
        **changes,
    ) -> Entity:
        if key is None:
            key = self.key
        try:
            duplicator = super().duplicate
        except AttributeError:
            duplicator = type(self)
        return duplicator(
            id=id,
            key=key,
            name=name,
            **changes,
        )


Entity = TypeVar("Entity", bound=_Entity)



# -*- coding: utf-8 -*-
"""
lories.core.register
~~~~~~~~~~~~~~~~~~~~


"""

from .registration import (  # noqa: F401
    Registration,
    RegistrationError,
)

from .registry import Registry  # noqa: F401

from .registrator import Registrator  # noqa: F401

from .context import RegistratorContext  # noqa: F401

from .access import RegistratorAccess  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories._core.typing
~~~~~~~~~~~~~~~~~~~


"""

import datetime as dt
from typing import TypeVar

import pandas as pd
import pytz as tz

Timestamp = TypeVar("Timestamp", pd.Timestamp, dt.datetime)
Timezone = TypeVar("Timezone", tz.BaseTzInfo, dt.tzinfo)

__all__ = [
    "Timestamp",
    "Timezone",
]



# -*- coding: utf-8 -*-
"""
lories.connectors.context
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from concurrent import futures
from concurrent.futures import Future
from typing import Callable, Dict, Optional, Type

from lories._core._channels import Channels  # noqa
from lories._core._connector import Connector, _Connector, _ConnectorContext  # noqa
from lories.connectors.errors import ConnectorError
from lories.connectors.tasks import ConnectTask
from lories.core.configs import Configurations
from lories.core.register import RegistratorContext, Registry
from lories.data.channels import ChannelState

registry = Registry[_Connector]()


# noinspection PyShadowingBuiltins
def register_connector_type(
    type: str,
    *alias: str,
    factory: Callable[[ConnectorContext | Connector, Optional[Configurations]], Connector] = None,
    replace: bool = False,
) -> Callable[[Type[Connector]], Type[Connector]]:
    # noinspection PyShadowingNames
    def _register(cls: Type[Connector]) -> Type[Connector]:
        registry.register(cls, type, *alias, factory=factory, replace=replace)
        return cls

    return _register


# noinspection PyProtectedMember
class ConnectorContext(_ConnectorContext, RegistratorContext[Connector]):
    @property
    def _registry(self) -> Registry[Connector]:
        return registry

    # noinspection PyShadowingBuiltins
    def connect(
        self,
        filter: Optional[Callable[[Connector], bool]] = None,
        channels: Optional[Channels] = None,
        timeout: Optional[float] = None,
    ) -> None:
        self._connect(*self.filter(filter), channels=channels, timeout=timeout)

    def _connect(
        self,
        *connectors: Connector,
        channels: Optional[Channels] = None,
        timeout: Optional[float] = None,
    ) -> None:
        connect_futures = {}
        for connector in connectors:
            if not connector.is_enabled():
                self._logger.debug(
                    f"Skipping to connect disabled {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            if connector._is_connected():
                self._logger.debug(
                    f"Skipping already connected {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            if not connector._is_connectable():
                self._logger.debug(
                    f"Skipping not connectable {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            connect_task = self.__connect(connector, channels)
            connect_future = self._executor.submit(connect_task)
            connect_futures[connect_future] = connect_task

        self.__connect_futures(connect_futures, timeout)

    # noinspection PyUnresolvedReferences
    def __connect(self, connector: Connector, channels: Optional[Channels] = None) -> ConnectTask:
        self._logger.debug(f"Connecting {type(connector).__name__} '{connector.name}': {connector.id}")
        if channels is None:
            channels = self.context.filter(lambda c: c.has_connector(connector.id))
            channels.update(self.context.filter(lambda c: c.has_logger(connector.id)).apply(lambda c: c.from_logger()))

        return ConnectTask(connector, channels)

    def __connect_futures(
        self,
        tasks: Dict[Future, ConnectTask],
        timeout: Optional[float] = None,
    ) -> None:
        try:
            for future in futures.as_completed(tasks, timeout=timeout):
                tasks.pop(future)
                self.__connect_callback(future)

        except TimeoutError:
            for future, task in tasks.items():
                self._logger.warning(f"Timed out opening connector '{task.connector.id}' after {timeout} seconds")
                future.cancel()

    def __connect_callback(self, future: Future) -> None:
        try:
            connector = future.result()
            self._logger.info(f"Connected {type(connector).__name__} '{connector.name}': {connector.id}")

        except ConnectorError as e:
            self._logger.warning(f"Failed opening connector '{e.connector.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)

    # noinspection PyShadowingBuiltins
    def reconnect(
        self,
        filter: Optional[Callable[[Connector], bool]] = None,
    ) -> None:
        self._reconnect(*self.filter(filter))

    def _reconnect(self, *connectors: Connector) -> None:
        for connector in connectors:
            if not connector.is_enabled():
                self._logger.debug(
                    f"Skipping reconnecting disabled {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue

            if not connector._is_connected() and connector._connected:
                # Connection aborted and not yet disconnected properly
                self._disconnect(connector)
                continue

            connect_task = self.__connect(connector)
            connect_future = self._executor.submit(connect_task)
            connect_future.add_done_callback(self.__connect_callback)

    # noinspection PyShadowingBuiltins
    def disconnect(
        self,
        filter: Optional[Callable[[Connector], bool]] = None,
    ) -> None:
        self._disconnect(*self.filter(filter))

    def _disconnect(self, *connectors: Connector) -> None:
        for connector in reversed(connectors):
            if not connector._is_connected():
                self._logger.debug(
                    f"Skipping to disconnect unconnected {type(connector).__name__} '{connector.name}': {connector.id}"
                )
                continue
            self.__disconnect(connector)

    def __disconnect(self, connector: Connector) -> None:
        try:
            self._logger.debug(f"Disconnecting {type(connector).__name__} '{connector.name}': {connector.id}")
            connector.set_channels(ChannelState.DISCONNECTING)
            connector.disconnect()

            self._logger.info(f"Disconnected {type(connector).__name__} '{connector.name}': {connector.id}")

        except Exception as e:
            self._logger.warning(f"Failed closing connector '{connector.id}': {str(e)}")
            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.exception(e)
        finally:
            connector.set_channels(ChannelState.DISCONNECTED)
